# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint contact"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


def get_analysis_types():
    return "all"  # write for all analysis types


def get_sets_name():
    return "constraints_contact_surface_sets"


def get_constraint_title():
    return "Contact Constraints"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_meshdata_constraint(f, femobj, contact_obj, ccxwriter):
    # slave DEP
    f.write(f"*SURFACE, NAME=DEP{contact_obj.Name}\n")
    for refs, surf, is_sub_el in femobj["ContactSlaveFaces"]:
        if is_sub_el:
            for elem, fno in surf:
                f.write(f"{elem},S{fno}\n")
        else:
            for elem in surf:
                f.write(f"{elem},S2\n")

    # master IND
    f.write(f"*SURFACE, NAME=IND{contact_obj.Name}\n")
    for refs, surf, is_sub_el in femobj["ContactMasterFaces"]:
        if is_sub_el:
            for elem, fno in surf:
                f.write(f"{elem},S{fno}\n")
        else:
            for elem in surf:
                f.write(f"{elem},S2\n")


def write_constraint(f, femobj, contact_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module
    adjust = ""
    if contact_obj.Adjust.Value > 0:
        adjust = ", ADJUST={:.13G}".format(contact_obj.Adjust.getValueAs("mm").Value)

    f.write(
        "*CONTACT PAIR, INTERACTION=INT{}, TYPE=SURFACE TO SURFACE{}\n".format(
            contact_obj.Name, adjust
        )
    )
    ind_surf = "IND" + contact_obj.Name
    dep_surf = "DEP" + contact_obj.Name
    f.write(f"{dep_surf}, {ind_surf}\n")
    f.write(f"*SURFACE INTERACTION, NAME=INT{contact_obj.Name}\n")
    if contact_obj.SurfaceBehavior == "Linear":
        f.write("*SURFACE BEHAVIOR, PRESSURE-OVERCLOSURE=LINEAR\n")
        slope = contact_obj.Slope.getValueAs("MPa/mm").Value
        f.write(f"{slope:.13G}\n")
    elif contact_obj.SurfaceBehavior == "Hard":
        f.write("*SURFACE BEHAVIOR, PRESSURE-OVERCLOSURE=HARD\n")
    elif contact_obj.SurfaceBehavior == "Tied":
        f.write("*SURFACE BEHAVIOR, PRESSURE-OVERCLOSURE=TIED\n")
        slope = contact_obj.Slope.getValueAs("MPa/mm").Value
        f.write(f"{slope:.13G}\n")
    else:
        return
    if contact_obj.Friction:
        f.write("*FRICTION\n")
        friction = contact_obj.FrictionCoefficient
        stick = contact_obj.StickSlope.getValueAs("MPa/mm").Value
        f.write(f"{friction:.13G}, {stick:.13G}\n")
    if contact_obj.EnableThermalContact:
        f.write("*GAP CONDUCTANCE\n")
        for value in contact_obj.ThermalContactConductance:
            f.write(f"{value}\n")
        f.write("\n")



# Example how to use the basic sketcher tools
from Sketcher import *
from Part import *
from FreeCAD import *
import FreeCAD as App

# set some constances for the constraints
StartPoint = 1
EndPoint = 2
MiddlePoint = 3

# create a document and a Sketch object
if App.activeDocument() is None:
    App.newDocument()

f = App.activeDocument().addObject("Sketcher::SketchObject", "Sketch")

# add geometry to the sketch
f.Geometry = [
    LineSegment(Vector(0, 0, 0), Vector(2, 20, 0)),
    LineSegment(Vector(0, 0, 0), Vector(20, 2, 0)),
]

# add constraints to the sketch
f.Constraints = [Constraint("Vertical", 0), Constraint("Horizontal", 1)]

# recompute (solving) the sketch
App.activeDocument().recompute()

# add another constraint to tie the start points together
l = f.Constraints
l.append(Constraint("Coincident", 0, StartPoint, 1, StartPoint))
f.Constraints = l

# again recompute
App.activeDocument().recompute()

f.Geometry



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions for general geometrical calculations."""
## @package geometry
# \ingroup draftgeoutils
# \brief Provides various functions for general geometrical calculations.

import math
import lazy_loader.lazy_loader as lz

import FreeCAD as App
import DraftVecUtils

import draftutils.gui_utils as gui_utils

from draftgeoutils.general import geomType, vec

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def findPerpendicular(point, edgeslist, force=None):
    """Find the perpendicular distance between a point and a list of edges.

    If force is specified, only the edge[force] will be considered,
    and it will be considered infinite.

    Returns
    -------
    [vector_from_point_to_closest_edge, edge_index]
        The vector and the index in the list.

    None
        If no perpendicular vector could be found.
    """
    if not isinstance(edgeslist, list):
        try:
            edgeslist = edgeslist.Edges
        except AttributeError:
            print("Doesn't have 'Edges'")
            return None

    if force is None:
        valid = None
        for edge in edgeslist:
            dist = findDistance(point, edge, strict=True)
            if dist:
                if not valid:
                    valid = [dist, edgeslist.index(edge)]
                else:
                    if dist.Length < valid[0].Length:
                        valid = [dist, edgeslist.index(edge)]
        return valid
    else:
        edge = edgeslist[force]
        dist = findDistance(point, edge)
        if dist:
            return [dist, force]
        else:
            return None
        return None


def findDistance(point, edge, strict=False):
    """Return a vector from the point to its closest point on the edge.

    If `strict` is `True`, the vector will be returned
    only if its endpoint lies on the `edge`.
    Edge can also be a list of 2 points.
    """
    if isinstance(point, App.Vector):
        if isinstance(edge, list):
            segment = edge[1].sub(edge[0])
            chord = edge[0].sub(point)
            norm = segment.cross(chord)
            perp = segment.cross(norm)
            dist = DraftVecUtils.project(chord, perp)

            if not dist:
                return None

            newpoint = point.add(dist)

            if dist.Length == 0:
                return None

            if strict:
                s1 = newpoint.sub(edge[0])
                s2 = newpoint.sub(edge[1])
                if s1.Length <= segment.Length and s2.Length <= segment.Length:
                    return dist
                else:
                    return None
            else:
                return dist

        elif geomType(edge) == "Line":
            segment = vec(edge)
            chord = edge.Vertexes[0].Point.sub(point)
            norm = segment.cross(chord)
            perp = segment.cross(norm)
            dist = DraftVecUtils.project(chord, perp)

            if not dist:
                return None

            newpoint = point.add(dist)

            if dist.Length == 0:
                return None

            if strict:
                s1 = newpoint.sub(edge.Vertexes[0].Point)
                s2 = newpoint.sub(edge.Vertexes[-1].Point)
                if s1.Length <= segment.Length and s2.Length <= segment.Length:
                    return dist
                else:
                    return None
            else:
                return dist

        elif geomType(edge) == "Circle":
            ve1 = edge.Vertexes[0].Point
            if len(edge.Vertexes) > 1:
                ve2 = edge.Vertexes[-1].Point
            else:
                ve2 = None
            center = edge.Curve.Center
            segment = center.sub(point)

            if segment.Length == 0:
                return None

            ratio = (segment.Length - edge.Curve.Radius) / segment.Length
            dist = segment.multiply(ratio)
            newpoint = App.Vector.add(point, dist)

            if dist.Length == 0:
                return None

            if strict and ve2:
                # Note 1: DraftVecUtils.angle(App.Vector(1, 1, 0)) => -0.7854
                # Note 2: Angles are in the +pi to -pi range.
                ang1 = DraftVecUtils.angle(ve1.sub(center))
                ang2 = DraftVecUtils.angle(ve2.sub(center))
                angpt = DraftVecUtils.angle(newpoint.sub(center))
                if ang1 >= ang2:  # Arc does not cross the 9 o'clock point.
                    if ang1 >= angpt and angpt >= ang2:
                        return dist
                    else:
                        return None
                elif ang1 >= angpt or angpt >= ang2:
                    return dist
                else:
                    return None
            else:
                return dist

        elif geomType(edge) == "BSplineCurve" or geomType(edge) == "BezierCurve":
            try:
                pr = edge.Curve.parameter(point)
                np = edge.Curve.value(pr)
                dist = np.sub(point)
            except Part.OCCError:
                print("DraftGeomUtils: Unable to get curve parameter " "for point ", point)
                return None
            else:
                return dist
        else:
            print("DraftGeomUtils: Couldn't project point")
            return None
    else:
        print("DraftGeomUtils: Couldn't project point")
        return None


def get_spline_normal(edge, tol=-1):
    """Find the normal of a BSpline edge."""

    if edge.isNull():
        return None

    if is_straight_line(shape, tol):
        return None

    plane = edge.findPlane(tol)
    if plane:
        normal = plane.Axis
        return normal
    else:
        return None


def get_shape_normal(shape):
    """Find the normal of a shape or list of points or colinear edges, if possible."""

    # New function based on get_normal() drafted by @Roy_043
    # in discussion https://forum.freecad.org/viewtopic.php?p=717862#p717862
    #
    # The normal would not be affected  by the 3D view direction and flipped as
    # get_normal would be. Placement of the Shape is taken into account in this
    # new function instead :
    # - Normal of a planar shape is usually bi-directional, this function return
    #   the one 'in the direction' of the shape's placement 'normal' (z-direction).
    #   As reference, shape.findPlane() favour positive axes, get_normal()
    #   favour the 'view direction' (obtained by getViewDirection() ).
    # - Even when the Shape is an edge or colinear edges, its Placement is taken
    #   into account, and this function return one 'in the direction' of the
    #   shape's placement 'normal' (z-direction).
    #   https://forum.freecad.org/viewtopic.php?p=715850#p715850
    # The normal direction of a Draft wire (co-planar) or arc used to depends on
    # the way the shape is constructed i.e. by vertexes in clockwise or
    # anti-clockwise.  This cryptic behaviour no longer matters.

    if shape.isNull():
        return None

    # Roy_043's remarks on the behavior of Shape.findPlane():
    # - https://forum.freecad.org/viewtopic.php?p=713993#p713993
    # Check if the shape is planar
    # If True: check if the outer wire is closed:
    #    if True  :   Return a (plane with) normal based on the direction of the outer wire (CW or CCW).
    #    if False : ? Return a (plane with) normal that points towards positive global axes with a certain priority
    #                 (probably? Z, Y and then X).
    # If False: return None.
    #
    # Further remarks :
    # - Tested Sketch with 2 colinear edges return a Plane with Axis
    # - Tested Sketch with 1 edge return no Plane with no Axis

    shape_rot = shape.Placement.Rotation
    plane = shape.findPlane()

    if plane is None:
        if not is_straight_line(shape):
            return None
        start_edge = shape.Edges[0]
        x_vec = start_edge.tangentAt(
            start_edge.FirstParameter
        )  # Return vector is in global coordinate
        local_x_vec = shape_rot.inverted().multVec(x_vec)  #
        local_rot = App.Rotation(local_x_vec, App.Vector(0, 1, 0), App.Vector(0, 0, 1), "XZY")
        # see https://blog.freecad.org/2023/01/16/the-rotation-api-in-freecad/ for App.Rotation(vecx, vecy, vecz, string)
        # discussion - https://forum.freecad.org/viewtopic.php?p=717738#p717738
        return shape_rot.multiply(local_rot).multVec(App.Vector(0, 0, 1))

    normal = plane.Axis
    shape_normal = shape_rot.multVec(App.Vector(0, 0, 1))
    # Now, check the normal direction of the plane (plane.Axis).
    # The final deduced normal should be 'in the direction' of the z-direction of the shape's placement,
    # if plane.Axis is in the 'opposite' direction of the Z direction of the shape's placement, reverse it.
    if normal.getAngle(shape_normal) > math.pi / 2:
        normal = normal.negative()
    return normal


def get_normal(shape, tol=-1):
    """Find the normal of a shape or list of points, if possible."""

    # for points
    if isinstance(shape, (list, tuple)):
        if len(shape) <= 2:
            return None
        else:
            poly = Part.makePolygon(shape)
            if is_straight_line(poly, tol):
                return None

            plane = poly.findPlane(tol)
            if plane:
                normal = plane.Axis
                return normal
            else:
                return None

    # for shapes
    if shape.isNull():
        return None

    if is_straight_line(shape, tol):
        return None
    else:
        plane = find_plane(shape, tol)
        if plane:
            normal = plane.Axis
        else:
            return None

    # Check the 3D view to flip the normal if the GUI is available
    if App.GuiUp:
        view = gui_utils.get_3d_view()
        if view is not None:
            v_dir = view.getViewDirection()
            if normal.getAngle(v_dir) < 0.78:
                normal = normal.negative()

    return normal


def getRotation(v1, v2=App.Vector(0, 0, 1)):
    """Get the rotation Quaternion between 2 vectors."""
    if (v1.dot(v2) > 0.999999) or (v1.dot(v2) < -0.999999):
        # vectors are opposite
        return None

    axis = v1.cross(v2)
    axis.normalize()
    # angle = math.degrees(math.sqrt(v1.Length^2 * v2.Length^2) + v1.dot(v2))
    angle = math.degrees(DraftVecUtils.angle(v1, v2, axis))
    return App.Rotation(axis, angle)


def is_planar(shape, tol=-1):
    """Return True if the given shape or list of points is planar."""

    # for points
    if isinstance(shape, list):
        if len(shape) <= 3:
            return True
        else:
            poly = Part.makePolygon(shape)
            if is_straight_line(poly, tol):
                return True

            plane = poly.findPlane(tol)
            if plane:
                return True
            else:
                return False

    # for shapes
    if shape.isNull():
        return False

    # because Part.Shape.findPlane return None for Vertex and straight edges
    if shape.ShapeType == "Vertex":
        return True

    if is_straight_line(shape, tol):
        return True

    plane = find_plane(shape, tol)
    if plane:
        return True
    else:
        return False


def is_straight_line(shape, tol=-1):
    """Return True if shape is a straight line.
    function used in other methods because Part.Shape.findPlane assign a
    plane and normal to straight wires creating privileged directions
    and to deal with straight wires with overlapped edges."""

    if shape.isNull():
        return False

    if len(shape.Faces) != 0:
        return False

    if len(shape.Edges) == 0:
        return False

    if len(shape.Edges) >= 1:
        start_edge = shape.Edges[0]
        dir_start_edge = start_edge.tangentAt(start_edge.FirstParameter)
        point_start_edge = start_edge.firstVertex().Point

        # set tolerance
        if tol <= 0:
            err = shape.globalTolerance(tol)
        else:
            err = tol

        for edge in shape.Edges:
            first_point = edge.firstVertex().Point
            last_point = edge.lastVertex().Point
            dir_edge = edge.tangentAt(edge.FirstParameter)
            # check if edge is curve or no parallel to start_edge
            # because sin(x) = x + O(x**3), for small angular deflection it's
            # enough use the cross product of directions (or dot with a normal)
            if (
                abs(edge.Length - first_point.distanceToPoint(last_point)) > err
                # https://forum.freecad.org/viewtopic.php?p=726101#p726101
                # Shape with parallel edges but not colinear used to return True
                # by this function, below line added fixes this bug.
                # Further remark on the below fix :
                # - get_normal() use this function, may had created problems
                #   previously but not reported; on the other hand, this fix
                #   seems have no 'regression' created as get_normal use
                #   plane.Axis (with 3DView check).  See get_shape_normal()
                #   which take into account shape's placement
                # - other functions had been using this also : Keep In View to
                #    see if there is 'regression' :
                #
                # $ grep -rni "is_straight" ./
                # ./draftfunctions/upgrade.py
                # ./draftgeoutils/geometry.py
                # ./draftmake/make_wire.py
                or first_point.distanceToLine(point_start_edge, dir_start_edge) > err  #
                or dir_start_edge.cross(dir_edge).Length > err
            ):
                return False

    return True


def are_coplanar(shape_a, shape_b, tol=-1):
    """Return True if exist a plane containing both shapes."""

    if shape_a.isNull() or shape_b.isNull():
        return False

    if not is_planar(shape_a, tol) or not is_planar(shape_b, tol):
        return False

    if shape_a.isEqual(shape_b):
        return True

    plane_a = find_plane(shape_a, tol)
    plane_b = find_plane(shape_b, tol)

    # set tolerance
    if tol <= 0:
        err = 1e-7
    else:
        err = tol

    if plane_a and plane_b:
        normal_a = plane_a.Axis
        normal_b = plane_b.Axis
        proj = plane_a.projectPoint(plane_b.Position)
        if normal_a.cross(normal_b).Length > err or plane_b.Position.sub(proj).Length > err:
            return False
        else:
            return True

    elif plane_a and not plane_b:
        normal_a = plane_a.Axis
        for vertex in shape_b.Vertexes:
            dir_ver_b = vertex.Point.sub(plane_a.Position).normalize()
            if abs(normal_a.dot(dir_ver_b)) > err:
                proj = plane_a.projectPoint(vertex.Point)
                if vertex.Point.sub(proj).Length > err:
                    return False
        return True

    elif plane_b and not plane_a:
        normal_b = plane_b.Axis
        for vertex in shape_a.Vertexes:
            dir_ver_a = vertex.Point.sub(plane_b.Position).normalize()
            if abs(normal_b.dot(dir_ver_a)) > err:
                proj = plane_b.projectPoint(vertex.Point)
                if vertex.Point.sub(proj).Length > err:
                    return False
        return True
    # not normal_a and not normal_b:
    else:
        points_a = [vertex.Point for vertex in shape_a.Vertexes]
        points_b = [vertex.Point for vertex in shape_b.Vertexes]
        poly = Part.makePolygon(points_a + points_b)
        if is_planar(poly, tol):
            return True
        else:
            return False


def get_spline_surface_normal(shape, tol=-1):
    """Check if shape formed by BSpline surfaces is planar and get normal.
    If shape is not planar return None."""

    if shape.isNull():
        return None

    if len(shape.Faces) == 0:
        return None

    # set tolerance
    if tol <= 0:
        err = shape.globalTolerance(tol)
    else:
        err = tol

    first_surf = shape.Faces[0].Surface

    if not first_surf.isPlanar(tol):
        return None

    # find bounds of first_surf
    u0, u1, v0, v1 = first_surf.bounds()
    u = (u0 + u1) / 2
    v = (v0 + v1) / 2
    first_normal = first_surf.normal(u, v)
    # check if all faces are planar and parallel
    for face in shape.Faces:
        surf = face.Surface
        if not surf.isPlanar(tol):
            return None
        u0, u1, v0, v1 = surf.bounds()
        u = (u0 + u1) / 2
        v = (v0 + v1) / 2
        surf_normal = surf.normal(u, v)
        if first_normal.cross(surf_normal).Length > err:
            return None

    normal = first_normal

    return normal


def find_plane(shape, tol=-1):
    """Find the plane containing the shape if possible.
    Use this function as a workaround due Part.Shape.findPlane
    fail to find plane on BSpline surfaces."""

    if shape.isNull():
        return None

    if shape.ShapeType == "Vertex":
        return None

    if is_straight_line(shape, tol):
        return None

    plane = shape.findPlane(tol)
    if plane:
        return plane
    elif len(shape.Faces) >= 1:
        # in case shape have BSpline surfaces
        normal = get_spline_surface_normal(shape, tol)
        if normal:
            position = shape.CenterOfMass
            return Part.Plane(position, normal)
        else:
            return None
    else:
        return None


def calculatePlacement(shape):
    """Return a placement located in the center of gravity of the shape.

    If the given shape is planar, return a placement located at the center
    of gravity of the shape, and oriented towards the shape's normal.
    Otherwise, it returns a null placement.
    """
    if not is_planar(shape):
        return App.Placement()

    pos = shape.BoundBox.Center
    norm = get_normal(shape)
    # for backward compatibility with previous getNormal implementation
    if norm is None:
        norm = App.Vector(0, 0, 1)
    pla = App.Placement()
    pla.Base = pos
    r = getRotation(norm)

    if r:
        pla.Rotation = r

    return pla


def mirror(point, edge):
    """Find mirror point relative to an edge."""
    normPoint = point.add(findDistance(point, edge, False))

    if normPoint:
        normPoint_point = App.Vector.sub(point, normPoint)
        normPoint_refl = normPoint_point.negative()
        refl = App.Vector.add(normPoint, normPoint_refl)
        return refl
    else:
        return None


def mirror_matrix(mtx, pos, nor):
    """Return a mirrored copy of a matrix.

    Parameters
    ----------
    mtx: Base::Matrix
        Matrix.
    pos: Base::Vector3
        Point on mirror plane.
    nor: Base::Vector3
        Normal of mirror plane.

    Returns
    -------
    Base::Matrix
    """
    # Code by Jolbas:
    # https://forum.freecad.org/viewtopic.php?p=702793#p702793
    mtx_copy = App.Matrix(mtx)
    mtx_copy.move(-pos)
    mtx_copy.scale(-1)
    mtx_copy = App.Rotation(nor, 180) * mtx_copy
    mtx_copy.move(pos)
    return mtx_copy


def uv_vectors_from_face(face, vec_z=App.Vector(0, 0, 1), tol=-1):
    """Return the u and v vectors of a planar face.

    It is up to the calling function to ensure the face is planar.

    If the u vector matches +/-vec_z, or the v vector matches -vec_z, the
    vectors are rotated to ensure the v vector matches +vec_z.

    Parameters
    ----------
    face: Part.Face
        Face.
    vec_z: Base::Vector3, optional
        Defaults to Vector(0, 0, 1).
        Z axis vector used for reference.
        Is replaced by Vector(0, 0, 1) if it matches the +/-normal of the face.
    tol: float, optional
        Defaults to -1.
        Internal tolerance. 1e-7 is used if tol <=0.

    Returns
    -------
    tuple
        U and v vector (Base::Vector3).
    """
    err = 1e-7 if tol <= 0 else tol
    if not vec_z.isEqual(App.Vector(0, 0, 1), err):
        nor = face.normalAt(0, 0)
        if vec_z.isEqual(nor, err) or vec_z.isEqual(nor.negative(), err):
            vec_z = App.Vector(0, 0, 1)
    vec_u, vec_v = face.tangentAt(0, 0)
    if face.Orientation == "Reversed":
        vec_u, vec_v = vec_v, vec_u
    if vec_v.isEqual(vec_z.negative(), err):
        vec_u, vec_v = vec_u.negative(), vec_v.negative()
    elif vec_u.isEqual(vec_z, err):
        vec_u, vec_v = vec_v.negative(), vec_u
    elif vec_u.isEqual(vec_z.negative(), err):
        vec_u, vec_v = vec_v, vec_u.negative()
    return vec_u, vec_v


def placement_from_face(face, vec_z=App.Vector(0, 0, 1), rotated=False, tol=-1):
    """Return a placement from the center of gravity, and the u and v vectors of a planar face.

    It is up to the calling function to ensure the face is planar.

    Parameters
    ----------
    face: Part.Face
        Face.
    vec_z: Base::Vector3, optional
        Defaults to Vector(0, 0, 1).
        Z axis vector used for reference.
        Is replaced by Vector(0, 0, 1) if it matches the +/-normal of the face.
    rotated: bool, optional
        Defaults to `False`.
        If `False` the v vector of the face defines the Y axis of the placement.
        If `True` the -v vector of the face defines the Z axis of the placement
        (used by Arch_Window).
        The u vector defines the X axis in both cases.
    tol: float, optional
        Defaults to -1.
        Internal tolerance. 1e-7 is used if tol <=0.

    Returns
    -------
    Base::Placement

    See also
    --------
    DraftGeomUtils.uv_vectors_from_face
    """
    pt_pos = face.CenterOfGravity
    vec_u, vec_v = uv_vectors_from_face(face, vec_z, tol)
    if rotated:
        return App.Placement(pt_pos, App.Rotation(vec_u, App.Vector(), vec_v.negative(), "XZY"))
    else:
        return App.Placement(pt_pos, App.Rotation(vec_u, vec_v, App.Vector(), "XYZ"))


def placement_from_points(pt_pos, pt_x, pt_y, as_vectors=False, tol=-1):
    """Return a placement from 3 points defining an origin, an X axis and a Y axis.

    If the vectors calculated from the arguments are too short or parallel,
    the returned placement will have a default rotation.

    Parameters
    ----------
    pt_pos: Base::Vector3
        Origin (Base of Placement).
    pt_x: Base::Vector3
        Point on positive X axis. Or X axis vector if as_vectors is `True`.
    pt_y: Base::Vector3
        Point on positive Y axis. Or Y axis vector if as_vectors is `True`.
    as_vectors: bool, optional
        Defaults to `False`.
        If `True` treat pt_x and pt_y as axis vectors.
    tol: float, optional
        Defaults to -1.
        Internal tolerance. 1e-7 is used if tol <=0.

    Returns
    -------
    Base::Placement

    See also
    --------
    DraftGeomUtils.getRotation
    DraftVecUtils.getRotation
    """
    err = 1e-7 if tol <= 0 else tol
    if as_vectors is False:
        vec_u = pt_x - pt_pos
        vec_v = pt_y - pt_pos
    else:
        vec_u = App.Vector(pt_x)
        vec_v = App.Vector(pt_y)

    if vec_u.Length < err or vec_v.Length < err:
        rot = App.Rotation()
    else:
        vec_u.normalize()
        vec_v.normalize()
        if vec_u.isEqual(vec_v, err) or vec_u.isEqual(vec_v.negative(), err):
            rot = App.Rotation()
        else:
            rot = App.Rotation(vec_u, vec_v, App.Vector(), "XYZ")

    return App.Placement(pt_pos, rot)


# Code separated from WorkingPlane.py (offsetToPoint function).
# Note that the return value of this function has the opposite sign.
def distance_to_plane(point, base, normal):
    """Return the signed distance from a plane to a point.

    The distance is positive if the point lies on the +normal side of the plane.

    Parameters
    ----------
    point: Base::Vector3
        Point to project.
    base: Base::Vector3
        Point on plane.
    normal: Base::Vector3
        Normal of plane.

    Returns
    -------
    float
    """
    return (point - base).dot(normal)


# Code separated from WorkingPlane.py (projectPoint function).
# See: https://github.com/FreeCAD/FreeCAD/pull/5307
def project_point_on_plane(point, base, normal, direction=None, force_projection=False, tol=-1):
    """Project a point onto a plane.

    Parameters
    ----------
    point: Base::Vector3
        Point to project.
    base: Base::Vector3
        Point on plane.
    normal: Base::Vector3
        Normal of plane.
    direction: Base::Vector3, optional
        Defaults to `None` in which case the normal is used.
        Direction of projection.
    force_projection: Bool, optional
        Defaults to `False`.
        If `True` forces the projection if the deviation between the direction
        and the normal is less than tol from the orthogonality. The direction
        of projection is then modified to a tol deviation between the direction
        and the orthogonal.
    tol: float, optional
        Defaults to -1.
        Internal tolerance. 1e-7 is used if tol <=0.

    Returns
    -------
    Base::Vector3 or `None`
    """
    err = 1e-7 if tol <= 0 else tol
    normal = App.Vector(normal).normalize()
    if direction is None:
        direction = normal
    else:
        direction = App.Vector(direction).normalize()

    cos = direction.dot(normal)
    delta_ax_proj = (point - base).dot(normal)
    # check the only conflicting case: direction orthogonal to normal
    if abs(cos) < err:
        if force_projection:
            cos = math.copysign(err, delta_ax_proj)
            direction = normal.cross(direction).cross(normal) - cos * normal
        else:
            return None

    return point - delta_ax_proj / cos * direction


# compatibility layer

getSplineNormal = get_spline_normal

getNormal = get_normal

isPlanar = is_planar


## @}



# ***************************************************************************
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

# to run the example use:
"""
from femexamples.material_nl_platewithhole import setup
setup()
"""

import FreeCAD
from FreeCAD import Vector as vec

import Part
from Part import makeCircle as ci
from Part import makeLine as ln

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Material NL Plate with Hole",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["fixed", "force"],
        "solvers": ["ccxtools"],
        "material": "nonlinear",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.material_nl_platewithhole import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=24&t=31997&start=30
https://forum.freecad.org/viewtopic.php?t=33974&start=90
https://forum.freecad.org/viewtopic.php?t=35893
https://forum.freecad.org/viewtopic.php?t=23101

Nonlinear material example, plate with hole.

plate: 400x200x10 mm
hole: diameter 100 mm (half cross section)
load: 130 MPa tension
linear material: Steel, E = 210000 MPa, my = 0.3
nonlinear material: '240.0, 0.0' to '270.0, 0.025'
TODO nonlinear material: give more information, use values from harry
TODO compare results with example from HarryvL

"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric object
    v1 = vec(-200, -100, 0)
    v2 = vec(200, -100, 0)
    v3 = vec(200, 100, 0)
    v4 = vec(-200, 100, 0)
    l1 = ln(v1, v2)
    l2 = ln(v2, v3)
    l3 = ln(v3, v4)
    l4 = ln(v4, v1)
    v5 = vec(0, 0, 0)
    c1 = ci(50, v5)
    face = Part.makeFace([Part.Wire([l1, l2, l3, l4]), c1], "Part::FaceMakerBullseye")
    geom_obj = doc.addObject("Part::Feature", "Hole_Plate")
    geom_obj.Shape = face.extrude(vec(0, 0, 10))
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.SplitInputWriter = False
        solver_obj.AnalysisType = "static"
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
        solver_obj.GeometricalNonlinearity = "nonlinear"
        solver_obj.MaterialNonlinearity = "nonlinear"
    analysis.addObject(solver_obj)

    # linear material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "Material_lin")
    matprop = material_obj.Material
    matprop["Name"] = "CalculiX-Steel"
    matprop["YoungsModulus"] = "210000 MPa"
    matprop["PoissonRatio"] = "0.30"
    material_obj.Material = matprop
    analysis.addObject(material_obj)

    # nonlinear material
    name_nlm = "Material_nonlin"
    nonlinear_mat = ObjectsFem.makeMaterialMechanicalNonlinear(doc, material_obj, name_nlm)
    nonlinear_mat.YieldPoints = ["240.0, 0.0", "270.0, 0.025"]
    analysis.addObject(nonlinear_mat)
    # check solver attributes, Nonlinearity needs to be set to nonlinear

    # constraint fixed
    con_fixed = ObjectsFem.makeConstraintFixed(doc, "ConstraintFixed")
    con_fixed.References = [(geom_obj, "Face4")]
    analysis.addObject(con_fixed)

    # pressure constraint
    con_pressure = ObjectsFem.makeConstraintPressure(doc, "ConstraintPressure")
    con_pressure.References = [(geom_obj, "Face2")]
    con_pressure.Pressure = "130.0 MPa"
    con_pressure.Reversed = True
    analysis.addObject(con_pressure)

    # mesh
    from .meshes.mesh_platewithhole_tetra10 import create_nodes, create_elements

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import pathlib
import tempfile
import FreeCAD
import Path
from PySide import QtGui, QtCore

translate = FreeCAD.Qt.translate


def _is_writable_dir(path: pathlib.Path) -> bool:
    """
    Check if a path is a writable directory.
    Returns True if writable, False otherwise.
    """
    if not path.is_dir():
        return False
    try:
        with tempfile.NamedTemporaryFile(dir=str(path), delete=True):
            return True
    except (OSError, PermissionError):
        return False


class AssetPreferencesPage:
    def __init__(self, parent=None):
        self.form = QtGui.QToolBox()
        self.form.setWindowTitle(translate("CAM_PreferencesAssets", "Assets"))

        asset_path_widget = QtGui.QWidget()
        main_layout = QtGui.QHBoxLayout(asset_path_widget)

        # Create widgets
        self.asset_path_label = QtGui.QLabel(translate("CAM_PreferencesAssets", "Asset Directory:"))
        self.asset_path_edit = QtGui.QLineEdit()
        self.asset_path_note_label = QtGui.QLabel(
            translate(
                "CAM_PreferencesAssets",
                "Note: Select the directory that will contain the "
                "Tool folder with Bit/, Shape/, and Library/ subfolders.",
            )
        )
        self.asset_path_note_label.setWordWrap(True)
        self.select_path_button = QtGui.QToolButton()
        self.select_path_button.setIcon(QtGui.QIcon.fromTheme("folder-open"))
        self.select_path_button.clicked.connect(self.selectAssetPath)
        self.reset_path_button = QtGui.QPushButton(translate("CAM_PreferencesAssets", "Reset"))
        self.reset_path_button.clicked.connect(self.resetAssetPath)

        # Set note label font to italic
        font = self.asset_path_note_label.font()
        font.setItalic(True)
        self.asset_path_note_label.setFont(font)

        # Layout for asset path section
        edit_button_layout = QtGui.QGridLayout()
        edit_button_layout.addWidget(self.asset_path_label, 0, 0, QtCore.Qt.AlignVCenter)
        edit_button_layout.addWidget(self.asset_path_edit, 0, 1, QtCore.Qt.AlignVCenter)
        edit_button_layout.addWidget(self.select_path_button, 0, 2, QtCore.Qt.AlignVCenter)
        edit_button_layout.addWidget(self.reset_path_button, 0, 3, QtCore.Qt.AlignVCenter)
        edit_button_layout.addWidget(self.asset_path_note_label, 1, 1, 1, 1, QtCore.Qt.AlignTop)
        edit_button_layout.addItem(
            QtGui.QSpacerItem(0, 0, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding),
            2,
            0,
            1,
            4,
        )

        main_layout.addLayout(edit_button_layout, QtCore.Qt.AlignTop)

        self.form.addItem(asset_path_widget, translate("CAM_PreferencesAssets", "Assets"))

    def selectAssetPath(self):
        # Implement directory selection dialog
        path = QtGui.QFileDialog.getExistingDirectory(
            self.form,
            translate("CAM_PreferencesAssets", "Select Asset Directory"),
            self.asset_path_edit.text(),
        )
        if path:
            self.asset_path_edit.setText(str(path))

    def resetAssetPath(self):
        # Implement resetting path to default
        default_path = Path.Preferences.getDefaultAssetPath()
        self.asset_path_edit.setText(str(default_path))

    def saveSettings(self):
        # Check path is writable, then call Path.Preferences.setAssetPath()
        asset_path = pathlib.Path(self.asset_path_edit.text())
        param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/CAM")
        if param.GetBool("CheckAssetPathWritable", True):
            if not _is_writable_dir(asset_path):
                QtGui.QMessageBox.warning(
                    self.form,
                    translate("CAM_PreferencesAssets", "Warning"),
                    translate("CAM_PreferencesAssets", "The selected asset path is not writable."),
                )
                return False
        Path.Preferences.setAssetPath(asset_path)
        Path.Preferences.setLastToolLibrary("")
        return True

    def loadSettings(self):
        # Get the raw preference value, not the versioned path
        pref = Path.Preferences.tool_preferences()
        asset_path = pref.GetString(Path.Preferences.ToolPath, "")
        if not asset_path:
            asset_path = str(Path.Preferences.getDefaultAssetPath())
        self.asset_path_edit.setText(asset_path)



# ***************************************************************************
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM Elasticity Elmer writer"
__author__ = "Uwe Stöhr"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from FreeCAD import Console
from FreeCAD import Units

from .. import sifio
from .. import writer as general_writer
from femtools import femutils
from . import elasticity


class ElasticityWriter:

    def __init__(self, writer, solver):
        self.write = writer
        self.solver = solver

    def getElasticitySolver(self, equation):
        s = self.write.createLinearSolver(equation)
        # check if we need to update the equation
        self._updateElasticitySolver(equation)
        # output the equation parameters
        s["Equation"] = "Stress Solver"  # equation.Name
        s["Procedure"] = sifio.FileAttr("StressSolve/StressSolver")
        if equation.CalculateStrains is True:
            s["Calculate Strains"] = equation.CalculateStrains
        if equation.CalculateStresses is True:
            s["Calculate Stresses"] = equation.CalculateStresses
        if equation.CalculatePrincipal is True:
            s["Calculate Principal"] = equation.CalculatePrincipal
        if equation.CalculatePangle is True:
            s["Calculate Pangle"] = equation.CalculatePangle
        if equation.ConstantBulkSystem is True:
            s["Constant Bulk System"] = equation.ConstantBulkSystem
        s["Displace mesh"] = equation.DisplaceMesh
        s["Eigen Analysis"] = equation.EigenAnalysis
        if equation.EigenAnalysis is True:
            s["Eigen System Convergence Tolerance"] = equation.EigenSystemTolerance
            s["Eigen System Complex"] = equation.EigenSystemComplex
            if equation.EigenSystemComputeResiduals is True:
                s["Eigen System Compute Residuals"] = equation.EigenSystemComputeResiduals
            s["Eigen System Damped"] = equation.EigenSystemDamped
            s["Eigen System Max Iterations"] = equation.EigenSystemMaxIterations
            s["Eigen System Select"] = equation.EigenSystemSelect
            s["Eigen System Values"] = equation.EigenSystemValues
        if equation.FixDisplacement is True:
            s["Fix Displacement"] = equation.FixDisplacement
        s["Geometric Stiffness"] = equation.GeometricStiffness
        if equation.Incompressible is True:
            s["Incompressible"] = equation.Incompressible
        if equation.MaxwellMaterial is True:
            s["Maxwell Material"] = equation.MaxwellMaterial
        if equation.ModelLumping is True:
            s["Model Lumping"] = equation.ModelLumping
        if equation.ModelLumping is True:
            s["Model Lumping Filename"] = equation.ModelLumpingFilename
        s["Optimize Bandwidth"] = True
        if equation.StabilityAnalysis is True:
            s["Stability Analysis"] = equation.StabilityAnalysis
        s["Stabilize"] = equation.Stabilize
        if equation.UpdateTransientSystem is True:
            s["Update Transient System"] = equation.UpdateTransientSystem
        s["Variable"] = equation.Variable
        s["Variable DOFs"] = 3
        return s

    def handleElasticityEquation(self, bodies, equation):
        for b in bodies:
            # not for bodies with fluid material
            if not self.write.isBodyMaterialFluid(b):
                if equation.PlaneStress:
                    self.write.equation(b, "Plane Stress", equation.PlaneStress)

    def _updateElasticitySolver(self, equation):
        # updates older Elasticity equations
        if not hasattr(equation, "Variable"):
            equation.addProperty(
                "App::PropertyString",
                "Variable",
                "Elasticity",
                (
                    "Only change this if 'Incompressible' is set to true\n"
                    "according to the Elmer manual."
                ),
                locked=True,
            )
            equation.Variable = "Displacement"
        if hasattr(equation, "Bubbles"):
            # Bubbles was removed because it is unused by Elmer for the stress solver
            equation.removeProperty("Bubbles")
        if not hasattr(equation, "ConstantBulkSystem"):
            equation.addProperty(
                "App::PropertyBool",
                "ConstantBulkSystem",
                "Elasticity",
                "See Elmer manual for info",
                locked=True,
            )
        if not hasattr(equation, "DisplaceMesh"):
            equation.addProperty(
                "App::PropertyBool",
                "DisplaceMesh",
                "Elasticity",
                (
                    "If mesh is deformed by displacement field.\n"
                    "Set to False for 'Eigen Analysis'."
                ),
                locked=True,
            )
            # DisplaceMesh is true except if DoFrequencyAnalysis is true
            equation.DisplaceMesh = True
            if hasattr(equation, "DoFrequencyAnalysis"):
                if equation.DoFrequencyAnalysis is True:
                    equation.DisplaceMesh = False
        if not hasattr(equation, "EigenAnalysis"):
            # DoFrequencyAnalysis was renamed to EigenAnalysis
            # to follow the Elmer manual
            equation.addProperty(
                "App::PropertyBool",
                "EigenAnalysis",
                "Eigen Values",
                "If true, modal analysis",
                locked=True,
            )
            if hasattr(equation, "DoFrequencyAnalysis"):
                equation.EigenAnalysis = equation.DoFrequencyAnalysis
                equation.removeProperty("DoFrequencyAnalysis")
        if not hasattr(equation, "EigenSystemComplex"):
            equation.addProperty(
                "App::PropertyBool",
                "EigenSystemComplex",
                "Eigen Values",
                (
                    "Should be true if eigen system is complex\n"
                    "Must be false for a damped eigen value analysis."
                ),
                locked=True,
            )
            equation.EigenSystemComplex = True
        if not hasattr(equation, "EigenSystemComputeResiduals"):
            equation.addProperty(
                "App::PropertyBool",
                "EigenSystemComputeResiduals",
                "Eigen Values",
                "Computes residuals of eigen value system",
                locked=True,
            )
        if not hasattr(equation, "EigenSystemDamped"):
            equation.addProperty(
                "App::PropertyBool",
                "EigenSystemDamped",
                "Eigen Values",
                (
                    "Set a damped eigen analysis. Can only be\n"
                    "used if 'Linear Solver Type' is 'Iterative'."
                ),
                locked=True,
            )
        if not hasattr(equation, "EigenSystemMaxIterations"):
            equation.addProperty(
                "App::PropertyIntegerConstraint",
                "EigenSystemMaxIterations",
                "Eigen Values",
                "Max iterations for iterative eigensystem solver",
                locked=True,
            )
            equation.EigenSystemMaxIterations = (300, 1, int(1e8), 1)
        if not hasattr(equation, "EigenSystemSelect"):
            equation.addProperty(
                "App::PropertyEnumeration",
                "EigenSystemSelect",
                "Eigen Values",
                "Which eigenvalues are computed",
                locked=True,
            )
            equation.EigenSystemSelect = elasticity.EIGEN_SYSTEM_SELECT
            equation.EigenSystemSelect = "Smallest Magnitude"
        if not hasattr(equation, "EigenSystemTolerance"):
            equation.addProperty(
                "App::PropertyFloat",
                "EigenSystemTolerance",
                "Eigen Values",
                (
                    "Convergence tolerance for iterative eigensystem solve\n"
                    "Default is 100 times the 'Linear Tolerance'"
                ),
                locked=True,
            )
            equation.setExpression("EigenSystemTolerance", str(100 * equation.LinearTolerance))
        if not hasattr(equation, "EigenSystemValues"):
            # EigenmodesCount was renamed to EigenSystemValues
            # to follow the Elmer manual
            equation.addProperty(
                "App::PropertyInteger",
                "EigenSystemValues",
                "Eigen Values",
                "Number of lowest eigen modes",
                locked=True,
            )
            if hasattr(equation, "EigenmodesCount"):
                equation.EigenSystemValues = equation.EigenmodesCount
                equation.removeProperty("EigenmodesCount")
        if not hasattr(equation, "FixDisplacement"):
            equation.addProperty(
                "App::PropertyBool",
                "FixDisplacement",
                "Elasticity",
                "If displacements or forces are set,\nthereby model lumping is used",
                locked=True,
            )
        if not hasattr(equation, "GeometricStiffness"):
            equation.addProperty(
                "App::PropertyBool",
                "GeometricStiffness",
                "Elasticity",
                "Consider geometric stiffness",
                locked=True,
            )
        if not hasattr(equation, "Incompressible"):
            equation.addProperty(
                "App::PropertyBool",
                "Incompressible",
                "Elasticity",
                (
                    "Computation of incompressible material in connection\n"
                    "with viscoelastic Maxwell material and a custom 'Variable'"
                ),
                locked=True,
            )
        if not hasattr(equation, "MaxwellMaterial"):
            equation.addProperty(
                "App::PropertyBool",
                "MaxwellMaterial",
                "Elasticity",
                "Compute viscoelastic material model",
                locked=True,
            )
        if not hasattr(equation, "ModelLumping"):
            equation.addProperty(
                "App::PropertyBool", "ModelLumping", "Elasticity", "Use model lumping", locked=True
            )
        if not hasattr(equation, "ModelLumpingFilename"):
            equation.addProperty(
                "App::PropertyFile",
                "ModelLumpingFilename",
                "Elasticity",
                "File to save results from model lumping to",
                locked=True,
            )
        if not hasattr(equation, "PlaneStress"):
            equation.addProperty(
                "App::PropertyBool",
                "PlaneStress",
                "Equation",
                (
                    "Computes solution according to plane\nstress situation.\n"
                    "Applies only for 2D geometry."
                ),
                locked=True,
            )
        if not hasattr(equation, "StabilityAnalysis"):
            equation.addProperty(
                "App::PropertyBool",
                "StabilityAnalysis",
                "Elasticity",
                (
                    "If true, 'Eigen Analysis' is stability analysis.\n"
                    "Otherwise modal analysis is performed."
                ),
                locked=True,
            )
        if not hasattr(equation, "UpdateTransientSystem"):
            equation.addProperty(
                "App::PropertyBool",
                "UpdateTransientSystem",
                "Elasticity",
                "See Elmer manual for info",
                locked=True,
            )

    def handleElasticityConstants(self):
        pass

    def handleElasticityBndConditions(self):
        for obj in self.write.getMember("Fem::ConstraintPressure"):
            if obj.References:
                for name in obj.References[0][1]:
                    pressure = float(obj.Pressure.getValueAs("Pa"))
                    if not obj.Reversed:
                        pressure *= -1
                    self.write.boundary(name, "Normal Force", pressure)
                self.write.handled(obj)
        for obj in self.write.getMember("Fem::ConstraintFixed"):
            if obj.References:
                for name in obj.References[0][1]:
                    self.write.boundary(name, "Displacement 1", 0.0)
                    self.write.boundary(name, "Displacement 2", 0.0)
                    self.write.boundary(name, "Displacement 3", 0.0)
                self.write.handled(obj)
        for obj in self.write.getMember("Fem::ConstraintForce"):
            if obj.References:
                for name in obj.References[0][1]:
                    force = float(obj.Force.getValueAs("N"))
                    self.write.boundary(name, "Force 1", obj.DirectionVector.x * force)
                    self.write.boundary(name, "Force 2", obj.DirectionVector.y * force)
                    self.write.boundary(name, "Force 3", obj.DirectionVector.z * force)
                    self.write.boundary(name, "Force 1 Normalize by Area", True)
                    self.write.boundary(name, "Force 2 Normalize by Area", True)
                    self.write.boundary(name, "Force 3 Normalize by Area", True)
                self.write.handled(obj)
        for obj in self.write.getMember("Fem::ConstraintDisplacement"):
            if obj.References:
                for name in obj.References[0][1]:
                    if obj.useFlowSurfaceForce:
                        self.write.boundary(name, "FSI BC", obj.useFlowSurfaceForce)
                        # if useFlowSurfaceForce no displacements must be output
                        continue
                    if not obj.xFree:
                        if not obj.hasXFormula:
                            displacement = float(obj.xDisplacement.getValueAs("m"))
                        else:
                            displacement = obj.xDisplacementFormula
                        self.write.boundary(name, "Displacement 1", displacement)
                    if not obj.yFree:
                        if not obj.hasYFormula:
                            displacement = float(obj.yDisplacement.getValueAs("m"))
                        else:
                            displacement = obj.yDisplacementFormula
                        self.write.boundary(name, "Displacement 2", displacement)
                    if not obj.zFree:
                        if not obj.hasZFormula:
                            displacement = float(obj.zDisplacement.getValueAs("m"))
                        else:
                            displacement = obj.zDisplacementFormula
                        self.write.boundary(name, "Displacement 3", displacement)
                self.write.handled(obj)
        for obj in self.write.getMember("Fem::ConstraintSpring"):
            if obj.References:
                for name in obj.References[0][1]:
                    if obj.ElmerStiffness == "Normal Stiffness":
                        spring = float(obj.NormalStiffness.getValueAs("N/m"))
                    else:
                        spring = float(obj.TangentialStiffness.getValueAs("N/m"))
                    self.write.boundary(name, "Spring", spring)
                self.write.handled(obj)

    def handleElasticityInitial(self, bodies):
        pass

    def handleElasticityBodyForces(self, bodies):
        obj = self.write.getSingleMember("Fem::ConstraintSelfWeight")
        if obj is not None:
            for name in bodies:
                gravity = self.write.convert(obj.GravityAcceleration.toStr(), "L/T^2")
                if self.write.getBodyMaterial(name) is None:
                    raise general_writer.WriteError(
                        f"The body {name} is not referenced in any material.\n\n"
                    )
                m = self.write.getBodyMaterial(name).Material

                densityQuantity = Units.Quantity(m["Density"])
                dimension = "M/L^3"
                if name.startswith("Edge"):
                    # not tested, bernd
                    # TODO: test
                    densityQuantity.Unit = Units.Unit(-2, 1)
                    dimension = "M/L^2"
                density = self.write.convert(densityQuantity, dimension)

                force1 = gravity * obj.GravityDirection.x * density
                force2 = gravity * obj.GravityDirection.y * density
                force3 = gravity * obj.GravityDirection.z * density
                self.write.bodyForce(name, "Stress Bodyforce 1", force1)
                self.write.bodyForce(name, "Stress Bodyforce 2", force2)
                self.write.bodyForce(name, "Stress Bodyforce 3", force3)
            self.write.handled(obj)

    def handleElasticityMaterial(self, bodies):
        # density
        # is needed for self weight constraints and frequency analysis
        density_needed = False
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerElasticity"):
                if equation.EigenAnalysis is True:
                    density_needed = True
                    break  # there could be a second equation without frequency
        gravObj = self.write.getSingleMember("Fem::ConstraintSelfWeight")
        if gravObj is not None:
            density_needed = True
        # temperature
        tempObj = self.write.getSingleMember("Fem::ConstraintInitialTemperature")
        if tempObj is not None:
            refTemp = float(tempObj.initialTemperature.getValueAs("K"))
            for name in bodies:
                self.write.material(name, "Reference Temperature", refTemp)
        # get the material data for all bodies
        for obj in self.write.getMember("App::MaterialObject"):
            m = obj.Material
            refs = obj.References[0][1] if obj.References else self.write.getAllBodies()
            for name in (n for n in refs if n in bodies):
                # don't evaluate fluid material
                if self.write.isBodyMaterialFluid(name):
                    break
                if "YoungsModulus" not in m:
                    Console.PrintMessage(f"m: {m}\n")
                    # it is no fluid but also no solid
                    # -> user set no material reference at all
                    # that now material is known
                    raise general_writer.WriteError(
                        "There are two or more materials with empty references.\n\n"
                        "Set for the materials to what solid they belong to.\n"
                    )
                self.write.material(name, "Name", m["Name"])
                if density_needed is True:
                    self.write.material(name, "Density", self.write.getDensity(m))
                self.write.material(name, "Youngs Modulus", self._getYoungsModulus(m))
                self.write.material(name, "Poisson ratio", float(m["PoissonRatio"]))
                if tempObj:
                    self.write.material(
                        name,
                        "Heat expansion Coefficient",
                        self.write.convert(m["ThermalExpansionCoefficient"], "O^-1"),
                    )

    def _getYoungsModulus(self, m):
        youngsModulus = self.write.convert(m["YoungsModulus"], "M/(L*T^2)")
        if self.write.getMeshDimension() == 2:
            youngsModulus *= 1e3
        return youngsModulus


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "Tools for the work with CalculiX solver"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"


from PySide.QtCore import QProcess, QThread, QProcessEnvironment
import os
import shutil

from vtkmodules.util import numpy_support as vtk_np
from vtkmodules.vtkCommonDataModel import vtkMultiBlockDataSet
import numpy as np

import FreeCAD
import Fem

from . import writer
from .. import settings

from femmesh import meshsetsgetter
from femtools import membertools
from femtools.objecttools import ObjectTools


class CalculiXTools(ObjectTools):

    name = "CalculiX"

    def __init__(self, obj):
        super().__init__(obj)
        self.model_file = ""

    def prepare(self):
        from femtools.checksanalysis import check_member_for_solver_calculix

        self._clear_results()

        message = check_member_for_solver_calculix(
            self.analysis,
            self.obj,
            membertools.get_mesh_to_solve(self.analysis)[0],
            membertools.AnalysisMember(self.analysis),
        )

        mesh_obj = membertools.get_mesh_to_solve(self.analysis)[0]
        meshdatagetter = meshsetsgetter.MeshSetsGetter(
            self.analysis,
            self.obj,
            mesh_obj,
            membertools.AnalysisMember(self.analysis),
        )
        meshdatagetter.get_mesh_sets()

        # write solver input
        w = writer.FemInputWriterCcx(
            self.analysis,
            self.obj,
            mesh_obj,
            meshdatagetter.member,
            self.obj.WorkingDirectory,
            meshdatagetter.mat_geo_sets,
        )
        self.model_file = w.write_solver_input()
        # report to user if task succeeded
        self.input_deck = os.path.splitext(os.path.basename(self.model_file))[0]

    def compute(self):
        self._clear_results()
        ccx_bin = settings.get_binary("Calculix")
        env = QProcessEnvironment.systemEnvironment()
        num_cpu = self.fem_param.GetGroup("Ccx").GetInt(
            "AnalysisNumCPUs", QThread.idealThreadCount()
        )
        env.insert("OMP_NUM_THREADS", str(num_cpu))
        pastix_prec = "1" if self.obj.PastixMixedPrecision else "0"
        env.insert("PASTIX_MIXED_PRECISION", pastix_prec)
        self.process.setProcessEnvironment(env)
        self.process.setWorkingDirectory(self.obj.WorkingDirectory)

        command_list = ["-i", os.path.join(self.obj.WorkingDirectory, self.input_deck)]
        self.process.start(ccx_bin, command_list)

        return self.process

    def update_properties(self):
        # TODO at the moment, only one .vtm file is assumed
        self._load_vtk_results()
        self._load_dat_results()

    def _clear_results(self):
        # result is a 'Result.vtm' file and a 'Result' directory
        # with the .vtu files
        dir_content = os.listdir(self.obj.WorkingDirectory)
        for f in dir_content:
            path = os.path.join(self.obj.WorkingDirectory, f)
            base, ext = os.path.splitext(path)
            if ext == ".vtm":
                # remove .vtm file
                os.remove(path)
                # remove dir with .vtu files
                shutil.rmtree(base)
            elif ext == ".dat":
                # remove .dat file
                os.remove(path)

    def _load_dat_results(self):
        # search dat output
        keep_result = self.fem_param.GetGroup("General").GetBool("KeepResultsOnReRun", False)
        dat = None
        for res in self.obj.Results:
            if res.isDerivedFrom("App::TextDocument"):
                dat = res

        if not dat or keep_result:
            # create dat output
            dat = self.obj.Document.addObject("App::TextDocument", self.obj.Name + "Output")
            self.analysis.addObject(dat)
            tmp = self.obj.Results
            tmp.append(dat)
            self.obj.Results = tmp

        files = os.listdir(self.obj.WorkingDirectory)
        for f in files:
            if f.endswith(".dat"):
                dat_file = os.path.join(self.obj.WorkingDirectory, f)
                with open(dat_file, "r") as file:
                    dat.Text = file.read()
                break

    def _load_vtk_results(self):
        # search current pipeline
        keep_result = self.fem_param.GetGroup("General").GetBool("KeepResultsOnReRun", False)
        pipeline = None
        create = False
        for res in self.obj.Results:
            if res.isDerivedFrom("Fem::FemPostPipeline"):
                pipeline = res

        if not pipeline or keep_result:
            # create pipeline
            pipeline = self.obj.Document.addObject("Fem::FemPostPipeline", self.obj.Name + "Result")
            self.analysis.addObject(pipeline)
            tmp = self.obj.Results
            tmp.append(pipeline)
            self.obj.Results = tmp
            create = True

        frd_result_prefix = os.path.join(self.obj.WorkingDirectory, self.input_deck)
        binary_mode = self.fem_param.GetGroup("Ccx").GetBool("BinaryOutput", False)
        Fem.frdToVTK(frd_result_prefix + ".frd", binary_mode)
        files = os.listdir(self.obj.WorkingDirectory)
        for f in files:
            if f.endswith(".vtm"):
                vtm_file = os.path.join(self.obj.WorkingDirectory, f)
                pipeline.read(vtm_file)
                multi_block = pipeline.Data
                multi_block = self._generate_derived_result(multi_block)
                if getattr(self.obj, "DisplaceMesh", False):
                    multi_block = self._generate_disp_mesh(multi_block)
                pipeline.Data = multi_block
                break

        pipeline.renameArrays(self.frd_var_conversion(self.obj.AnalysisType))
        self._set_time_info(pipeline)

        if create and FreeCAD.GuiUp:
            # default display mode
            view_obj = pipeline.ViewObject
            view_obj.DisplayMode = "Surface"
            view_obj.SelectionStyle = "BoundBox"
            enum_field = view_obj.getEnumerationsOfProperty("Field")
            default_field = self._get_default_field()
            if default_field in enum_field:
                view_obj.Field = default_field

    def frd_var_conversion(self, analysis_type):
        common = {
            "CONTACT": "Contact Displacement",
            "PE": "Plastic Strain",
            "CELS": "Contact Energy",
            "ECD": "Current Density",
            "EMFB": "Magnetic Field",
            "EMFE": "Electric Field",
            "ENER": "Internal Energy Density",
            "DISP": "Displacement",
            "TOSTRAIN": "Strain",
            "STRESS": "Stress",
            "STR(%)": "Error",
        }
        thermo = {"FLUX": "Heat Flux", "T": "Temperature"}
        electrostatic = {"T": "Potential", "FLUX": "Electric Flux Density"}

        match analysis_type:
            case "thermomech":
                common.update(thermo)
            case "electromagnetic":
                common.update(electrostatic)

        return common

    def _get_default_field(self):
        match self.obj.AnalysisType:
            case "static" | "frequency" | "buckling":
                return "Displacement"
            case "thermomech":
                return "Temperature"
            case "electromagnetic":
                return "Potential"
            case _:
                return "None"

    def _generate_disp_mesh(self, mb):
        try:
            for i in range(mb.GetNumberOfBlocks()):
                grid = mb.GetBlock(i)
                if grid is None:
                    continue
                pd = grid.GetPointData()
                if (pd is None) or not pd.HasArray("DISP"):
                    continue
                points = grid.GetPoints()
                pos = points.GetData()
                disp = pd.GetAbstractArray("DISP")
                disp_points = vtk_np.vtk_to_numpy(pos) + vtk_np.vtk_to_numpy(disp)
                disp_points = vtk_np.numpy_to_vtk(disp_points)
                points.SetData(disp_points)

            mb_out = vtkMultiBlockDataSet()
            mb_out.DeepCopy(mb)

            return mb_out

        except Exception as e:
            # return original data
            FreeCAD.Console.PrintWarning("Displaced mesh not generated\n")
            return mb

    def _generate_derived_result(self, mb):
        try:
            for i in range(mb.GetNumberOfBlocks()):
                grid = mb.GetBlock(i)
                if grid is None:
                    continue
                pd = grid.GetPointData()
                if (pd is None) or not pd.HasArray("STRESS"):
                    continue

                stress = pd.GetAbstractArray("STRESS")
                s_xx, s_yy, s_zz, s_xy, s_yz, s_xz = vtk_np.vtk_to_numpy(stress).transpose()
                # create symmetric matrix from [Sxx, Syy, Szz, Sxy, Syz, Sxz]
                sym_stress = (
                    np.array([s_xx, s_xy, s_xz, s_xy, s_yy, s_yz, s_xz, s_yz, s_zz])
                    .transpose()
                    .reshape(-1, 3, 3)
                )
                prin_val, prin_axes = np.linalg.eigh(sym_stress)

                sigma_1 = prin_val[:, 2]
                sigma_3 = prin_val[:, 0]
                sigma_2 = prin_val[:, 1]
                sigma_1_axis = prin_axes[:, :, 2]
                sigma_3_axis = prin_axes[:, :, 0]
                sigma_2_axis = prin_axes[:, :, 1]

                sigma_1_vtk = vtk_np.numpy_to_vtk(sigma_1)
                sigma_1_vtk.SetName("Major Principal Stress")
                sigma_3_vtk = vtk_np.numpy_to_vtk(sigma_3)
                sigma_3_vtk.SetName("Minor Principal Stress")
                sigma_2_vtk = vtk_np.numpy_to_vtk(sigma_2)
                sigma_2_vtk.SetName("Intermediate Principal Stress")
                sigma_1_axis_vtk = vtk_np.numpy_to_vtk(sigma_1_axis)
                sigma_1_axis_vtk.SetName("Major Principal Stress Direction")
                sigma_3_axis_vtk = vtk_np.numpy_to_vtk(sigma_3_axis)
                sigma_3_axis_vtk.SetName("Minor Principal Stress Direction")
                sigma_2_axis_vtk = vtk_np.numpy_to_vtk(sigma_2_axis)
                sigma_2_axis_vtk.SetName("Intermediate Principal Stress Direction")

                # add principal stress
                pd.AddArray(sigma_1_vtk)
                pd.AddArray(sigma_3_vtk)
                pd.AddArray(sigma_2_vtk)
                # add principal directions
                pd.AddArray(sigma_1_axis_vtk)
                pd.AddArray(sigma_3_axis_vtk)
                pd.AddArray(sigma_2_axis_vtk)

                sigma_diff = np.array([sigma_1 - sigma_2, sigma_2 - sigma_3, sigma_3 - sigma_1])

                # von Mises stress
                von_mises = np.linalg.norm(sigma_diff, axis=0) / np.sqrt(2)
                von_mises_vtk = vtk_np.numpy_to_vtk(von_mises)
                von_mises_vtk.SetName("von Mises Stress")
                pd.AddArray(von_mises_vtk)

                # max shear stress
                max_shear = 1 / 2 * np.max(np.abs(sigma_diff), axis=0)
                max_shear_vtk = vtk_np.numpy_to_vtk(max_shear)
                max_shear_vtk.SetName("Tresca Stress")
                pd.AddArray(max_shear_vtk)

                # max abs principal stress
                max_abs_sigma = np.max(np.abs(np.array([sigma_1, sigma_2, sigma_3])), axis=0)
                max_abs_sigma_vtk = vtk_np.numpy_to_vtk(max_abs_sigma)
                max_abs_sigma_vtk.SetName("Maximum Absolute Principal Stress")
                pd.AddArray(max_abs_sigma_vtk)

            mb_out = vtkMultiBlockDataSet()
            mb_out.DeepCopy(mb)

            return mb_out

        except Exception as e:
            # return original data
            FreeCAD.Console.PrintWarning("Derived result not generated\n")
            return mb

    def _set_time_info(self, pipeline):
        if not hasattr(pipeline, "setTimeInfo"):
            return
        match self.obj.AnalysisType:
            case "frequency":
                pipeline.setTimeInfo("Frequency", FreeCAD.Units.Frequency)
            case "buckling":
                pipeline.setTimeInfo("Buckling factor", FreeCAD.Units.Unit())
            case _:
                pipeline.setTimeInfo("TimeStep", FreeCAD.Units.TimeSpan)

    def version(self):
        p = QProcess()
        ccx_bin = settings.get_binary("Calculix")
        p.start(ccx_bin, ["-v"])
        p.waitForFinished()
        info = p.readAll().data().decode()
        return info



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2018 Kresimir Tusek <kresimir.tusek@gmail.com>          *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************

import FreeCADGui
import FreeCAD
import Path.Op.Adaptive as PathAdaptive
import Path.Op.Gui.Base as PathOpGui
import Path.Op.Gui.FeatureExtension as PathFeatureExtensionsGui
from PySide import QtCore

import Path.Base.Gui.Util as PathGuiUtil


class TaskPanelOpPage(PathOpGui.TaskPanelPage):
    def getForm(self):
        """getForm() ... return UI"""

        form = FreeCADGui.PySideUic.loadUi(":/panels/PageOpAdaptiveEdit.ui")
        comboToPropertyMap = [("Side", "Side"), ("OperationType", "OperationType")]

        enumTups = PathAdaptive.PathAdaptive.propertyEnumerations(dataType="raw")

        self.populateCombobox(form, enumTups, comboToPropertyMap)
        return form

    def initPage(self, obj):
        self.form.LiftDistance.setProperty("unit", obj.LiftDistance.getUserPreferred()[2])
        self.form.HelixDiameterLimit.setProperty(
            "unit", obj.HelixDiameterLimit.getUserPreferred()[2]
        )
        self.form.KeepToolDownRatio.setProperty("unit", obj.KeepToolDownRatio.getUserPreferred()[2])
        self.form.StockToLeave.setProperty("unit", obj.StockToLeave.getUserPreferred()[2])

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return list of signals for updating obj"""
        signals = []
        signals.append(self.form.Side.currentIndexChanged)
        signals.append(self.form.OperationType.currentIndexChanged)
        signals.append(self.form.toolController.currentIndexChanged)
        signals.append(self.form.stepOverPercent.valueChanged)
        signals.append(self.form.Tolerance.valueChanged)
        signals.append(self.form.HelixAngle.valueChanged)
        signals.append(self.form.HelixConeAngle.valueChanged)
        signals.append(self.form.HelixDiameterLimit.valueChanged)
        signals.append(self.form.LiftDistance.valueChanged)
        signals.append(self.form.KeepToolDownRatio.valueChanged)
        signals.append(self.form.StockToLeave.valueChanged)
        signals.append(self.form.coolantController.currentIndexChanged)
        if hasattr(self.form.ForceInsideOut, "checkStateChanged"):  # Qt version >= 6.7.0
            signals.append(self.form.ForceInsideOut.checkStateChanged)
            signals.append(self.form.FinishingProfile.checkStateChanged)
            signals.append(self.form.useOutline.checkStateChanged)
        else:  # Qt version < 6.7.0
            signals.append(self.form.ForceInsideOut.stateChanged)
            signals.append(self.form.FinishingProfile.stateChanged)
            signals.append(self.form.useOutline.stateChanged)
        signals.append(self.form.StopButton.toggled)
        return signals

    def setFields(self, obj):
        self.selectInComboBox(obj.Side, self.form.Side)
        self.selectInComboBox(obj.OperationType, self.form.OperationType)
        self.form.stepOverPercent.setValue(obj.StepOver)
        self.form.Tolerance.setValue(int(obj.Tolerance * 100))

        self.form.HelixAngle.setText(
            FreeCAD.Units.Quantity(obj.HelixAngle, FreeCAD.Units.Angle).UserString
        )

        self.form.HelixConeAngle.setText(
            FreeCAD.Units.Quantity(obj.HelixConeAngle, FreeCAD.Units.Angle).UserString
        )

        self.form.HelixDiameterLimit.setProperty("rawValue", obj.HelixDiameterLimit.Value)

        self.form.LiftDistance.setProperty("rawValue", obj.LiftDistance.Value)

        if hasattr(obj, "KeepToolDownRatio"):
            self.form.KeepToolDownRatio.setProperty("rawValue", obj.KeepToolDownRatio.Value)
            # self.form.KeepToolDownRatio.setValue(obj.KeepToolDownRatio)

        if hasattr(obj, "StockToLeave"):
            self.form.StockToLeave.setProperty("rawValue", obj.StockToLeave.Value)

        self.form.ForceInsideOut.setChecked(obj.ForceInsideOut)
        self.form.FinishingProfile.setChecked(obj.FinishingProfile)
        self.form.useOutline.setChecked(obj.UseOutline)
        self.setupToolController(obj, self.form.toolController)
        self.setupCoolant(obj, self.form.coolantController)
        self.form.StopButton.setChecked(obj.Stopped)
        obj.setEditorMode("AdaptiveInputState", 2)  # hide this property
        obj.setEditorMode("AdaptiveOutputState", 2)  # hide this property
        obj.setEditorMode("StopProcessing", 2)  # hide this property
        obj.setEditorMode("Stopped", 2)  # hide this property

    def getFields(self, obj):
        if obj.Side != str(self.form.Side.currentData()):
            obj.Side = str(self.form.Side.currentData())

        if obj.OperationType != str(self.form.OperationType.currentData()):
            obj.OperationType = str(self.form.OperationType.currentData())

        if obj.StepOver != self.form.stepOverPercent.value():
            obj.StepOver = self.form.stepOverPercent.value()

        obj.Tolerance = 1.0 * self.form.Tolerance.value() / 100.0
        PathGuiUtil.updateInputField(obj, "HelixAngle", self.form.HelixAngle)
        PathGuiUtil.updateInputField(obj, "HelixConeAngle", self.form.HelixConeAngle)
        PathGuiUtil.updateInputField(obj, "HelixDiameterLimit", self.form.HelixDiameterLimit)
        PathGuiUtil.updateInputField(obj, "LiftDistance", self.form.LiftDistance)

        if hasattr(obj, "KeepToolDownRatio"):
            PathGuiUtil.updateInputField(obj, "KeepToolDownRatio", self.form.KeepToolDownRatio)

        if hasattr(obj, "StockToLeave"):
            PathGuiUtil.updateInputField(obj, "StockToLeave", self.form.StockToLeave)

        obj.ForceInsideOut = self.form.ForceInsideOut.isChecked()
        obj.FinishingProfile = self.form.FinishingProfile.isChecked()
        obj.UseOutline = self.form.useOutline.isChecked()
        obj.Stopped = self.form.StopButton.isChecked()
        if obj.Stopped:
            self.form.StopButton.setChecked(False)  # reset the button
            obj.StopProcessing = True

        self.updateToolController(obj, self.form.toolController)
        self.updateCoolant(obj, self.form.coolantController)
        obj.setEditorMode("AdaptiveInputState", 2)  # hide this property
        obj.setEditorMode("AdaptiveOutputState", 2)  # hide this property
        obj.setEditorMode("StopProcessing", 2)  # hide this property
        obj.setEditorMode("Stopped", 2)  # hide this property

    def taskPanelBaseLocationPage(self, obj, features):
        if not hasattr(self, "extensionsPanel"):
            self.extensionsPanel = PathFeatureExtensionsGui.TaskPanelExtensionPage(obj, features)
        return self.extensionsPanel


Command = PathOpGui.SetupOperation(
    "Adaptive",
    PathAdaptive.Create,
    TaskPanelOpPage,
    "CAM_Adaptive",
    QtCore.QT_TRANSLATE_NOOP("CAM_Adaptive", "Adaptive"),
    QtCore.QT_TRANSLATE_NOOP("CAM_Adaptive", "Adaptive clearing and profiling"),
    PathAdaptive.SetupProperties,
)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2018 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Units
import Path
import argparse
import datetime
import shlex
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils
from builtins import open as pyopen

TOOLTIP = """
This is a postprocessor file for the Path workbench. It is used to
take a pseudo-G-code fragment outputted by a Path object, and output
real G-code suitable for a jtech photonics laser. This postprocessor, once placed
in the appropriate PathScripts folder, can be used directly from inside
FreeCAD, via the GUI importer or via python scripts with:

import jtech_post
jtech_post.export(object,"/path/to/file.ngc","")
"""

now = datetime.datetime.now()

parser = argparse.ArgumentParser(prog="jtech", add_help=False)
parser.add_argument("--no-header", action="store_true", help="suppress header output")
parser.add_argument("--no-comments", action="store_true", help="suppress comment output")
parser.add_argument("--line-numbers", action="store_true", help="prefix with line numbers")
parser.add_argument(
    "--no-show-editor",
    action="store_true",
    help="don't pop up editor before writing output",
)
parser.add_argument("--precision", default="3", help="number of digits of precision, default=3")
parser.add_argument(
    "--preamble",
    help='set commands to be issued before the first command, default="M05 S0\\nG90\\n"',
)
parser.add_argument(
    "--postamble",
    help='set commands to be issued after the last command, default="M05 S0\\nM2\\n"',
)
parser.add_argument(
    "--inches", action="store_true", help="Convert output for US imperial mode (G20)"
)
parser.add_argument(
    "--modal",
    action="store_true",
    help="Output the Same G-command Name USE NonModal Mode",
)
parser.add_argument("--axis-modal", action="store_true", help="Output the Same Axis Value Mode")
parser.add_argument(
    "--power-on-delay",
    default="255",
    help="milliseconds - Add a delay after laser on before moving to pre-heat material. Default=0",
)


TOOLTIP_ARGS = parser.format_help()

# These globals set common customization preferences
OUTPUT_COMMENTS = True
OUTPUT_HEADER = True
OUTPUT_LINE_NUMBERS = False
SHOW_EDITOR = True
MODAL = False  # if true commands are suppressed if the same as previous line.
OUTPUT_DOUBLES = True  # if false duplicate axis values are suppressed if the same as previous line.
COMMAND_SPACE = " "
LINENR = 100  # line number starting value

# These globals will be reflected in the Machine configuration of the project
UNITS = "G21"  # G21 for metric, G20 for us standard
UNIT_SPEED_FORMAT = "mm/min"
UNIT_FORMAT = "mm"

MACHINE_NAME = "JTECH Photonic Laser"
PRECISION = 3

# Preamble text will appear at the beginning of the GCODE output file.
PREAMBLE = """M05 S0
G90
"""

# Postamble text will appear following the last operation.
POSTAMBLE = """M05 S0
M2
"""

PRE_FEED = """M03
G4 P{}
"""

POST_FEED = """M05
"""

# Pre operation text will be inserted before every operation
PRE_OPERATION = """"""

# Post operation text will be inserted after every operation
POST_OPERATION = """"""

# Tool Change commands will be inserted before a tool change
TOOL_CHANGE = """"""

POWER_ON_DELAY = 0


def processArguments(argstring):
    global OUTPUT_HEADER
    global OUTPUT_COMMENTS
    global OUTPUT_LINE_NUMBERS
    global SHOW_EDITOR
    global PRECISION
    global PREAMBLE
    global POSTAMBLE
    global UNITS
    global UNIT_SPEED_FORMAT
    global UNIT_FORMAT
    global MODAL
    global OUTPUT_DOUBLES
    global POWER_ON_DELAY

    try:
        args = parser.parse_args(shlex.split(argstring))
        if args.no_header:
            OUTPUT_HEADER = False
        if args.no_comments:
            OUTPUT_COMMENTS = False
        if args.line_numbers:
            OUTPUT_LINE_NUMBERS = True
        if args.no_show_editor:
            SHOW_EDITOR = False
        print("Show editor = %d" % SHOW_EDITOR)
        PRECISION = args.precision
        if args.preamble is not None:
            PREAMBLE = args.preamble.replace("\\n", "\n")
        if args.postamble is not None:
            POSTAMBLE = args.postamble.replace("\\n", "\n")
        if args.inches:
            UNITS = "G20"
            UNIT_SPEED_FORMAT = "in/min"
            UNIT_FORMAT = "in"
            PRECISION = 4
        if args.modal:
            MODAL = True
        if args.axis_modal:
            OUTPUT_DOUBLES = False
        POWER_ON_DELAY = float(args.power_on_delay) / 1000  # milliseconds

    except Exception:
        return False

    return True


def export(objectslist, filename, argstring):
    if not processArguments(argstring):
        return None

    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print(
                "the object " + obj.Name + " is not a path. Please select only path and Compounds."
            )
            return None

    print("postprocessing...")
    gcode = ""

    # write header
    if OUTPUT_HEADER:
        gcode += linenumber() + "(Exported by FreeCAD)\n"
        gcode += linenumber() + "(Post Processor: " + __name__ + ")\n"
        gcode += linenumber() + "(Output Time:" + str(now) + ")\n"

    # Write the preamble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(begin preamble)\n"
    for line in PREAMBLE.splitlines():
        gcode += linenumber() + line + "\n"
    gcode += linenumber() + UNITS + "\n"

    for obj in objectslist:

        # do the pre_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(begin operation: %s)\n" % obj.Label
        for line in PRE_OPERATION.splitlines(True):
            gcode += linenumber() + line

        gcode += parse(obj)

        # do the post_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(finish operation: %s)\n" % obj.Label
        for line in POST_OPERATION.splitlines(True):
            gcode += linenumber() + line

    # do the post_amble
    if OUTPUT_COMMENTS:
        gcode += "(begin postamble)\n"
    for line in POSTAMBLE.splitlines():
        gcode += linenumber() + line + "\n"

    if FreeCAD.GuiUp and SHOW_EDITOR:
        dia = PostUtils.GCodeEditorDialog()
        dia.editor.setText(gcode)
        result = dia.exec_()
        if result:
            final = dia.editor.toPlainText()
        else:
            final = gcode
    else:
        final = gcode

    print("done postprocessing.")

    if not filename == "-":
        gfile = pyopen(filename, "wb")
        gfile.write(final)
        gfile.close()

    return final


def linenumber():
    global LINENR
    if OUTPUT_LINE_NUMBERS is True:
        LINENR += 10
        return "N" + str(LINENR) + " "
    return ""


def parse(pathobj):
    out = ""
    lastcommand = None
    precision_string = "." + str(PRECISION) + "f"
    currLocation = {}  # keep track for no doubles
    RAPID_MOVES = ["G0", "G00"]
    FEED_MOVES = ["G1", "G01", "G2", "G02", "G3", "G03"]
    # the order of parameters
    params = [
        "X",
        "Y",
        "Z",
        "A",
        "B",
        "C",
        "I",
        "J",
        "F",
        "S",
        "T",
        "Q",
        "R",
        "L",
        "H",
        "D",
        "P",
    ]
    firstmove = Path.Command("G0", {"X": -1, "Y": -1, "Z": -1, "F": 0.0})
    currLocation.update(firstmove.Parameters)  # set First location Parameters

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        for p in pathobj.Group:
            out += parse(p)
        return out
    else:  # parsing simple path

        # groups might contain non-path things like stock.
        if not hasattr(pathobj, "Path"):
            return out

        for c in PathUtils.getPathWithPlacement(pathobj).Commands:

            outstring = []
            command = c.Name

            controlstring = ""
            if command in FEED_MOVES and lastcommand in RAPID_MOVES:
                controlstring = PRE_FEED.format(POWER_ON_DELAY)
            elif command in RAPID_MOVES and lastcommand in FEED_MOVES:
                controlstring = POST_FEED
            if len(controlstring) > 0:
                out += controlstring

            outstring.append(command)

            # if modal: suppress the command if it is the same as the last one
            if MODAL is True:
                if command == lastcommand:
                    outstring.pop(0)

            if c.Name.startswith("(") and not OUTPUT_COMMENTS:  # command is a comment
                continue

            # Now add the remaining parameters in order
            for param in params:
                if param in c.Parameters:
                    if param == "F" and (
                        currLocation[param] != c.Parameters[param] or OUTPUT_DOUBLES
                    ):
                        if c.Name not in RAPID_MOVES:  # linuxcnc doesn't use rapid speeds
                            speed = Units.Quantity(c.Parameters["F"], FreeCAD.Units.Velocity)
                            if speed.getValueAs(UNIT_SPEED_FORMAT) > 0.0:
                                outstring.append(
                                    param
                                    + format(
                                        float(speed.getValueAs(UNIT_SPEED_FORMAT)),
                                        precision_string,
                                    )
                                )
                        else:
                            continue
                    elif param == "T":
                        outstring.append(param + str(int(c.Parameters["T"])))
                    elif param == "H":
                        outstring.append(param + str(int(c.Parameters["H"])))
                    elif param == "D":
                        outstring.append(param + str(int(c.Parameters["D"])))
                    elif param == "S":
                        outstring.append(param + str(int(c.Parameters["S"])))
                    else:
                        if (
                            (not OUTPUT_DOUBLES)
                            and (param in currLocation)
                            and (currLocation[param] == c.Parameters[param])
                        ):
                            continue
                        else:
                            pos = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)
                            outstring.append(
                                param + format(float(pos.getValueAs(UNIT_FORMAT)), precision_string)
                            )

            # store the latest command
            lastcommand = command
            currLocation.update(c.Parameters)

            # Check for Tool Change:
            if command == "M6":
                continue

            if command == "message":
                if OUTPUT_COMMENTS is False:
                    out = []
                else:
                    outstring.pop(0)  # remove the command

            # prepend a line number and append a newline
            if len(outstring) >= 1:
                if OUTPUT_LINE_NUMBERS:
                    outstring.insert(0, (linenumber()))

                # append the line to the final output
                for w in outstring:
                    out += w + COMMAND_SPACE
                out = out.strip() + "\n"

        return out


# print(__name__ + " gcode postprocessor loaded.")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2020 Werner Mayer <wmayer[at]users.sourceforge.net>     *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "BasicShapes.ViewProviderShapes"
__author__ = "Werner Mayer"
__url__ = "https://www.freecad.org"
__doc__ = "Basic shapes"


import os
import FreeCAD
import FreeCADGui

from PySide import QtGui


class ViewProviderTube:
    def __init__(self, obj):
        """Set this object to the proxy object of the actual view provider"""
        obj.Proxy = self
        obj.addExtension("PartGui::ViewProviderAttachExtensionPython")
        obj.setIgnoreOverlayIcon(True, "PartGui::ViewProviderAttachExtensionPython")

    def attach(self, obj):
        """Setup the scene sub-graph of the view provider, this method is mandatory"""
        return

    def setupContextMenu(self, viewObject, menu):
        action = menu.addAction(
            FreeCAD.Qt.translate("QObject", "Edit %1").replace("%1", viewObject.Object.Label)
        )
        action.triggered.connect(lambda: self.startDefaultEditMode(viewObject))
        return False

    def startDefaultEditMode(self, viewObject):
        document = viewObject.Document.Document
        if not document.HasPendingTransaction:
            text = FreeCAD.Qt.translate("QObject", "Edit %1").replace("%1", viewObject.Object.Label)
            document.openTransaction(text)
        viewObject.Document.setEdit(viewObject.Object, 0)

    def setEdit(self, viewObject, mode):
        if mode == 0:
            FreeCADGui.Control.showDialog(TaskTubeUI(viewObject))
            return True

    def unsetEdit(self, viewObject, mode):
        if mode == 0:
            FreeCADGui.Control.closeDialog()
            return True

    def getIcon(self):
        return ":/icons/parametric/Part_Tube_Parametric.svg"

    def dumps(self):
        return None

    def loads(self, state):
        return None


class TaskTubeUI:
    """A default task panel for editing tube objects."""

    def __init__(self, viewObject):
        self.viewObject = viewObject
        ui_file = os.path.join(os.path.dirname(__file__), "TaskTube.ui")
        ui = FreeCADGui.UiLoader()
        self.form = ui.load(ui_file)

        object = self.viewObject.Object
        self.form.tubeOuterRadius.setProperty("rawValue", object.OuterRadius.Value)
        self.form.tubeInnerRadius.setProperty("rawValue", object.InnerRadius.Value)
        self.form.tubeHeight.setProperty("rawValue", object.Height.Value)

        self.form.tubeOuterRadius.valueChanged.connect(lambda x: self.onChangeOuterRadius(x))
        self.form.tubeInnerRadius.valueChanged.connect(lambda x: self.onChangeInnerRadius(x))
        self.form.tubeHeight.valueChanged.connect(lambda x: self.onChangeHeight(x))

        FreeCADGui.ExpressionBinding(self.form.tubeOuterRadius).bind(object, "OuterRadius")
        FreeCADGui.ExpressionBinding(self.form.tubeInnerRadius).bind(object, "InnerRadius")
        FreeCADGui.ExpressionBinding(self.form.tubeHeight).bind(object, "Height")

    def onChangeOuterRadius(self, radius):
        object = self.viewObject.Object
        object.OuterRadius = radius
        object.recompute()

    def onChangeInnerRadius(self, radius):
        object = self.viewObject.Object
        object.InnerRadius = radius
        object.recompute()

    def onChangeHeight(self, height):
        object = self.viewObject.Object
        object.Height = height
        object.recompute()

    def accept(self):
        object = self.viewObject.Object
        if not object.isValid():
            QtGui.QMessageBox.warning(None, "Error", object.getStatusString())
            return False
        document = self.viewObject.Document.Document
        document.commitTransaction()
        document.recompute()
        self.viewObject.Document.resetEdit()
        return True

    def reject(self):
        guidocument = self.viewObject.Document
        document = guidocument.Document
        document.abortTransaction()
        document.recompute()
        guidocument.resetEdit()
        return True



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2010 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2010 Ken Cline <cline@frii.com>                         *
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides support functions to edit Draft objects.

All functions in this module work with Object coordinate space.
No conversion to global coordinate system is needed.

To support an new Object, Draft_Edit needs at least two functions:
getObjectPts(obj): returns a list of points on which Draft_Edit will display
    edit trackers
updateObject(obj, nodeIndex, v): update the give object according to the
    index of a moved edit tracker and the vector of the displacement
TODO: Abstract the code that handles the preview and move the object specific
    code to this module from main Draft_Edit module
"""
## @package gui_edit_draft_objects
# \ingroup draftguitools
# \brief Provides support functions to edit Draft objects.

__title__ = "FreeCAD Draft Edit Tool"
__author__ = (
    "Yorik van Havre, Werner Mayer, Martin Burbaum, Ken Cline, " "Dmitry Chigrin, Carlo Pavan"
)
__url__ = "https://www.freecad.org"

## \addtogroup draftguitools
# @{
import math
import FreeCAD as App
import FreeCADGui as Gui
import DraftVecUtils

from draftutils.translate import translate
import draftutils.utils as utils
from draftutils.messages import _err

import draftguitools.gui_trackers as trackers

from draftguitools.gui_edit_base_object import GuiTools


class DraftWireGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        for p in obj.Points:
            editpoints.append(p)
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        pts = obj.Points
        tol = 0.001  # TODO : Use default precision
        if node_idx == 0 and (v - pts[-1]).Length < tol:
            # DNC: user moved first point over last point -> Close curve
            obj.Closed = True
            pts[0] = v
            del pts[-1]
            obj.Points = pts
            return
        elif node_idx == len(pts) - 1 and (v - pts[0]).Length < tol:
            # DNC: user moved last point over first point -> Close curve
            obj.Closed = True
            del pts[-1]
            obj.Points = pts
            return
        elif v in pts:
            # DNC: checks if point enter is equal to other, this could cause a OCC problem
            _err = translate("draft", "This object does not support possible " "coincident points")
            App.Console.PrintMessage(_err + "\n")
            return

        # TODO: Make consistent operation with trackers and open wires
        # See: https://forum.freecad.org/viewtopic.php?f=23&t=56661
        # if obj.Closed:
        #    # DNC: project the new point to the plane of the face if present
        #    if hasattr(obj.Shape, "normalAt"):
        #        normal = obj.Shape.normalAt(0,0)
        #        point_on_plane = obj.Shape.Vertexes[0].Point
        #        v.projectToPlane(point_on_plane, normal)

        pts[node_idx] = v
        obj.Points = pts

    def get_edit_point_context_menu(self, edit_command, obj, node_idx):
        return [
            (translate("draft", "Delete Point"), lambda: self.delete_point(obj, node_idx)),
        ]

    def get_edit_obj_context_menu(self, edit_command, obj, position):
        return [
            (translate("draft", "Add Point"), lambda: self.add_point(edit_command, obj, position)),
            (self.get_open_close_menu_text(obj), lambda: self.open_close_wire(obj)),
            (self.get_reverse_menu_text(obj), lambda: self.reverse_wire(obj)),
        ]

    def get_open_close_menu_text(self, obj):
        """This function is overridden in the DraftBSplineGuiTools class."""
        if obj.Closed:
            return translate("draft", "Open Wire")
        else:
            return translate("draft", "Close Wire")

    def get_reverse_menu_text(self, obj):
        """This function is overridden in the DraftBSplineGuiTools class."""
        return translate("draft", "Reverse Wire")

    def init_preview_object(self, obj):
        return trackers.wireTracker(obj.Shape)

    def update_preview_object(self, edit_command, obj, node_idx, v):
        if utils.get_type(obj) in ["Wire"]:
            pointList = edit_command.globalize_vectors(obj, obj.Points)
            pointList[node_idx] = v
            if obj.Closed:
                pointList.append(pointList[0])
            edit_command.ghost.updateFromPointlist(pointList)

    def add_point(self, edit_command, obj, pos):
        """Add point to obj."""
        info, newPoint = edit_command.get_specific_object_info(obj, pos)

        if not info:
            return
        if not "Edge" in info["Component"]:
            return
        edgeIndex = int(info["Component"][4:])

        newPoints = []
        if hasattr(obj, "ChamferSize") and hasattr(obj, "FilletRadius"):
            # TODO: If Draft_Wire fails to calculate one of the fillets or chamfers
            #       this algo fails to identify the correct edge
            if obj.ChamferSize > 0 and obj.FilletRadius > 0:
                edgeIndex = (edgeIndex + 3) / 4
            elif obj.ChamferSize > 0 or obj.FilletRadius > 0:
                edgeIndex = (edgeIndex + 1) / 2

        for index, point in enumerate(obj.Points):
            if index == edgeIndex:
                newPoints.append(edit_command.localize_vector(obj, newPoint))
            newPoints.append(point)
        if obj.Closed and edgeIndex == len(obj.Points):
            # last segment when object is closed
            newPoints.append(edit_command.localize_vector(obj, newPoint))
        obj.Points = newPoints

        obj.recompute()

    def delete_point(self, obj, node_idx):
        if len(obj.Points) <= 2:
            _err(translate("draft", "Active object must have more than 2 points or nodes"))
            return

        pts = obj.Points
        pts.pop(node_idx)
        obj.Points = pts

        obj.recompute()

    def open_close_wire(self, obj):
        obj.Closed = not obj.Closed
        obj.recompute()

    def reverse_wire(self, obj):
        obj.Points = reversed(obj.Points)
        obj.recompute()


class DraftBSplineGuiTools(DraftWireGuiTools):

    def get_open_close_menu_text(self, obj):
        if obj.Closed:
            return translate("draft", "Open Spline")
        else:
            return translate("draft", "Close Spline")

    def get_reverse_menu_text(self, obj):
        return translate("draft", "Reverse Spline")

    def init_preview_object(self, obj):
        return trackers.bsplineTracker()

    def update_preview_object(self, edit_command, obj, node_idx, v):
        pointList = edit_command.globalize_vectors(obj, obj.Points)
        pointList[node_idx] = v
        if obj.Closed:
            pointList.append(pointList[0])
        edit_command.ghost.update(pointList)

    def add_point(self, edit_command, obj, pos):
        """Add point to obj."""
        info, global_pt = edit_command.get_specific_object_info(obj, pos)
        pt = edit_command.localize_vector(obj, global_pt)

        if not info or (pt is None):
            return

        pts = obj.Points
        if obj.Closed:
            curve = obj.Shape.Edges[0].Curve
        else:
            curve = obj.Shape.Curve
        uNewPoint = curve.parameter(obj.Placement.multVec(pt))
        uPoints = curve.getKnots()

        for i in range(len(uPoints) - 1):
            if (uNewPoint > uPoints[i]) and (uNewPoint < uPoints[i + 1]):
                pts.insert(i + 1, pt)
                break
        # DNC: fix: add points to last segment if curve is closed
        if obj.Closed and (uNewPoint > uPoints[-1]):
            pts.append(pt)
        obj.Points = pts

        obj.recompute()


class DraftRectangleGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        """Return the list of edipoints for the given Draft Rectangle.

        0 : Placement.Base
        1 : Length
        2 : Height
        """
        editpoints = []
        editpoints.append(App.Vector(0, 0, 0))
        editpoints.append(App.Vector(obj.Length, 0, 0))
        editpoints.append(App.Vector(0, obj.Height, 0))
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if node_idx == 0:
            obj.Placement.Base = obj.Placement.multVec(v)
        elif node_idx == 1:
            obj.Length = DraftVecUtils.project(v, App.Vector(1, 0, 0)).Length
        elif node_idx == 2:
            obj.Height = DraftVecUtils.project(v, App.Vector(0, 1, 0)).Length


class DraftCircleGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        """Return the list of edipoints for the given Draft Arc or Circle.

        circle:
        0 : Placement.Base or center
        1 : radius

        arc:
        0 : Placement.Base or center
        1 : first endpoint
        2 : second endpoint
        3 : midpoint
        """
        editpoints = []
        editpoints.append(App.Vector(0, 0, 0))
        if obj.FirstAngle == obj.LastAngle:
            # obj is a circle
            editpoints.append(App.Vector(obj.Radius, 0, 0))
        else:
            # obj is an arc
            editpoints.append(self.getArcStart(obj))  # First endpoint
            editpoints.append(self.getArcEnd(obj))  # Second endpoint
            editpoints.append(self.getArcMid(obj))  # Midpoint
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if obj.FirstAngle == obj.LastAngle:
            # object is a circle
            if node_idx == 0:
                obj.Placement.Base = obj.Placement.multVec(v)
            elif node_idx == 1:
                obj.Radius = v.Length

        else:
            # obj is an arc
            if alt_edit_mode == 0:
                import Part

                if node_idx == 0:
                    # center point
                    p1 = self.getArcStart(obj)
                    p2 = self.getArcEnd(obj)
                    p0 = DraftVecUtils.project(v, self.getArcMid(obj))
                    obj.Radius = p1.sub(p0).Length
                    obj.FirstAngle = -math.degrees(DraftVecUtils.angle(p1.sub(p0)))
                    obj.LastAngle = -math.degrees(DraftVecUtils.angle(p2.sub(p0)))
                    obj.Placement.Base = obj.Placement.multVec(p0)

                else:
                    """Edit arc by 3 points."""
                    v = obj.Placement.multVec(v)
                    p1 = obj.Placement.multVec(self.getArcStart(obj))
                    p2 = obj.Placement.multVec(self.getArcMid(obj))
                    p3 = obj.Placement.multVec(self.getArcEnd(obj))

                    if node_idx == 1:  # first point
                        p1 = v
                    elif node_idx == 3:  # midpoint
                        p2 = v
                    elif node_idx == 2:  # second point
                        p3 = v

                    arc = Part.ArcOfCircle(p1, p2, p3)
                    import Part

                    s = arc.toShape()
                    # Part.show(s) DEBUG
                    p0 = arc.Location
                    obj.Placement.Base = p0
                    obj.Radius = arc.Radius

                    delta = s.Vertexes[0].Point
                    obj.FirstAngle = -math.degrees(DraftVecUtils.angle(p1.sub(p0)))
                    delta = s.Vertexes[1].Point
                    obj.LastAngle = -math.degrees(DraftVecUtils.angle(p3.sub(p0)))

            elif alt_edit_mode == 1:
                # edit arc by center radius FirstAngle LastAngle
                if node_idx == 0:
                    obj.Placement.Base = obj.Placement.multVec(v)
                else:
                    dangle = math.degrees(math.atan2(v[1], v[0]))
                    if node_idx == 1:
                        obj.FirstAngle = dangle
                    elif node_idx == 2:
                        obj.LastAngle = dangle
                    elif node_idx == 3:
                        obj.Radius = v.Length

    def get_edit_point_context_menu(self, edit_command, obj, node_idx):
        actions = None
        if obj.FirstAngle != obj.LastAngle:
            if node_idx == 0:  # user is over arc start point
                return [
                    (
                        translate("draft", "Move Arc"),
                        lambda: self.handle_move_arc(edit_command, obj, node_idx),
                    ),
                ]
            elif node_idx == 1:  # user is over arc start point
                return [
                    (
                        translate("draft", "Set First Angle"),
                        lambda: self.handle_set_first_angle(edit_command, obj, node_idx),
                    ),
                ]
            elif node_idx == 2:  # user is over arc end point
                return [
                    (
                        translate("draft", "Set Last Angle"),
                        lambda: self.handle_set_last_angle(edit_command, obj, node_idx),
                    ),
                ]
            elif node_idx == 3:  # user is over arc mid point
                return [
                    (
                        translate("draft", "Set Radius"),
                        lambda: self.handle_set_radius(edit_command, obj, node_idx),
                    ),
                ]

    def handle_move_arc(self, edit_command, obj, node_idx):
        edit_command.alt_edit_mode = 1
        edit_command.startEditing(obj, node_idx)

    def handle_set_first_angle(self, edit_command, obj, node_idx):
        edit_command.alt_edit_mode = 1
        edit_command.startEditing(obj, node_idx)

    def handle_set_last_angle(self, edit_command, obj, node_idx):
        edit_command.alt_edit_mode = 1
        edit_command.startEditing(obj, node_idx)

    def handle_set_radius(self, edit_command, obj, node_idx):
        edit_command.alt_edit_mode = 1
        edit_command.startEditing(obj, node_idx)

    def get_edit_obj_context_menu(self, edit_command, obj, position):
        # Do not show the `Invert arc` option for circles:
        if obj.FirstAngle == obj.LastAngle:
            return

        return [
            (translate("draft", "Invert Arc"), lambda: self.arcInvert(obj)),
        ]

    def init_preview_object(self, obj):
        return trackers.arcTracker()

    def update_preview_object(self, edit_command, obj, node_idx, v):
        edit_command.ghost.setCenter(obj.getGlobalPlacement().Base)
        edit_command.ghost.setRadius(obj.Radius)
        if obj.FirstAngle == obj.LastAngle:
            # obj is a circle
            edit_command.ghost.circle = True
            if node_idx == 0:
                edit_command.ghost.setCenter(v)
            elif node_idx == 1:
                radius = v.sub(obj.getGlobalPlacement().Base).Length
                edit_command.ghost.setRadius(radius)
        else:
            if edit_command.alt_edit_mode == 0:
                # edit by 3 points
                if node_idx == 0:
                    # center point
                    p1 = edit_command.localize_vector(obj, obj.Shape.Vertexes[0].Point)
                    p2 = edit_command.localize_vector(obj, obj.Shape.Vertexes[1].Point)
                    p0 = DraftVecUtils.project(
                        edit_command.localize_vector(obj, v),
                        edit_command.localize_vector(
                            obj, (self.getArcMid(obj, global_placement=True))
                        ),
                    )
                    edit_command.ghost.autoinvert = False
                    edit_command.ghost.setRadius(p1.sub(p0).Length)
                    edit_command.ghost.setStartPoint(obj.Shape.Vertexes[1].Point)
                    edit_command.ghost.setEndPoint(obj.Shape.Vertexes[0].Point)
                    edit_command.ghost.setCenter(edit_command.globalize_vector(obj, p0))
                    return
                else:
                    p1 = self.getArcStart(obj, global_placement=True)
                    p2 = self.getArcMid(obj, global_placement=True)
                    p3 = self.getArcEnd(obj, global_placement=True)
                    if node_idx == 1:
                        p1 = v
                    elif node_idx == 3:
                        p2 = v
                    elif node_idx == 2:
                        p3 = v
                    edit_command.ghost.setBy3Points(p1, p2, p3)
            elif edit_command.alt_edit_mode == 1:
                # edit by center radius angles
                edit_command.ghost.setStartAngle(math.radians(obj.FirstAngle))
                edit_command.ghost.setEndAngle(math.radians(obj.LastAngle))
                if node_idx == 0:
                    edit_command.ghost.setCenter(v)
                elif node_idx == 1:
                    edit_command.ghost.setStartPoint(v)
                elif node_idx == 2:
                    edit_command.ghost.setEndPoint(v)
                elif node_idx == 3:
                    edit_command.ghost.setRadius(edit_command.localize_vector(obj, v).Length)

    def getArcStart(self, obj, global_placement=False):  # Returns object midpoint
        if utils.get_type(obj) == "Circle":
            return self.pointOnCircle(obj, obj.FirstAngle, global_placement)

    def getArcEnd(self, obj, global_placement=False):  # Returns object midpoint
        if utils.get_type(obj) == "Circle":
            return self.pointOnCircle(obj, obj.LastAngle, global_placement)

    def getArcMid(self, obj, global_placement=False):  # Returns object midpoint
        if utils.get_type(obj) == "Circle":
            if obj.LastAngle > obj.FirstAngle:
                midAngle = obj.FirstAngle + (obj.LastAngle - obj.FirstAngle) / 2.0
            else:
                midAngle = obj.FirstAngle + (obj.LastAngle - obj.FirstAngle) / 2.0
                midAngle += App.Units.Quantity(180, App.Units.Angle)
            return self.pointOnCircle(obj, midAngle, global_placement)

    def pointOnCircle(self, obj, angle, global_placement=False):
        if utils.get_type(obj) == "Circle":
            px = obj.Radius * math.cos(math.radians(angle))
            py = obj.Radius * math.sin(math.radians(angle))
            p = App.Vector(px, py, 0.0)
            if global_placement:
                p = obj.getGlobalPlacement().multVec(p)
            return p
        return None

    def arcInvert(self, obj):
        obj.FirstAngle, obj.LastAngle = obj.LastAngle, obj.FirstAngle
        obj.recompute()


class DraftEllipseGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        editpoints.append(App.Vector(0, 0, 0))
        editpoints.append(App.Vector(obj.MajorRadius, 0, 0))
        editpoints.append(App.Vector(0, obj.MinorRadius, 0))
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if node_idx == 0:
            obj.Placement.Base = obj.Placement.multVec(v)
        elif node_idx == 1:
            if v.Length >= obj.MinorRadius:
                obj.MajorRadius = v.Length
            else:
                obj.MajorRadius = obj.MinorRadius
        elif node_idx == 2:
            if v.Length <= obj.MajorRadius:
                obj.MinorRadius = v.Length
            else:
                obj.MinorRadius = obj.MajorRadius


class DraftPolygonGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        editpoints.append(App.Vector(0, 0, 0))
        if obj.DrawMode == "inscribed":
            editpoints.append(obj.Placement.inverse().multVec(obj.Shape.Vertexes[0].Point))
        else:
            editpoints.append(
                obj.Placement.inverse().multVec(
                    (obj.Shape.Vertexes[0].Point + obj.Shape.Vertexes[1].Point) / 2
                )
            )
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if node_idx == 0:
            obj.Placement.Base = obj.Placement.multVec(v)
        elif node_idx == 1:
            obj.Radius = v.Length
        obj.recompute()


class DraftDimensionGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        p = obj.ViewObject.Proxy.textpos.translation.getValue()
        editpoints.append(obj.Start)
        editpoints.append(obj.End)
        editpoints.append(obj.Dimline)
        editpoints.append(App.Vector(p[0], p[1], p[2]))
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if node_idx == 0:
            obj.Start = v
        elif node_idx == 1:
            obj.End = v
        elif node_idx == 2:
            obj.Dimline = v
        elif node_idx == 3:
            obj.ViewObject.TextPosition = v


class DraftLabelGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        if obj.StraightDirection != "Custom":
            editpoints.append(obj.Points[0])
            editpoints.append(obj.Points[1])
            editpoints.append(obj.Points[2])
            if editpoints[0].isEqual(editpoints[1], 0.001):
                # If StraightDistance == 0.0, editpoint 1 is exactly on editpoint 0 => Move slightly editpoint 1 to allow selection of editpoint 0
                editpoints[1] = 0.9 * editpoints[1] + 0.1 * editpoints[2]
        else:
            pass  # TODO : support custom direction with any number of points
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        if obj.StraightDirection != "Custom":
            if node_idx == 0:
                vector_p1_p2 = obj.Placement.inverse().multVec(
                    obj.Points[1]
                ) - obj.Placement.inverse().multVec(v)
                if obj.StraightDirection == "Horizontal":
                    if round(vector_p1_p2.y, utils.precision()) == 0.0:
                        obj.StraightDistance = vector_p1_p2.x
                elif obj.StraightDirection == "Vertical":
                    if round(vector_p1_p2.x, utils.precision()) == 0.0:
                        obj.StraightDistance = vector_p1_p2.y
                obj.Placement.Base = v
            elif node_idx == 1:
                vector_p1_p2 = obj.Placement.inverse().multVec(v) - obj.Placement.inverse().multVec(
                    obj.Placement.Base
                )
                if obj.StraightDirection == "Horizontal":
                    if round(vector_p1_p2.y, utils.precision()) == 0.0:
                        obj.StraightDistance = vector_p1_p2.x
                    else:
                        obj.Placement.Base = obj.Placement.Base + (v - obj.Points[1])
                elif obj.StraightDirection == "Vertical":
                    if round(vector_p1_p2.x, utils.precision()) == 0.0:
                        obj.StraightDistance = vector_p1_p2.y
                    else:
                        obj.Placement.Base = obj.Placement.Base + (v - obj.Points[1])
            elif node_idx == 2:
                obj.TargetPoint = v
        else:
            pass  # TODO : support custom direction with any number of points


class DraftBezCurveGuiTools(GuiTools):

    def __init__(self):
        pass

    def get_edit_points(self, obj):
        editpoints = []
        for p in obj.Points:
            editpoints.append(p)
        return editpoints

    def update_object_from_edit_points(self, obj, node_idx, v, alt_edit_mode=0):
        pts = obj.Points
        # DNC: check for coincident startpoint/endpoint to auto close the curve
        tol = 0.001
        if (
            ((node_idx == 0) and ((v - pts[-1]).Length < tol))
            or (node_idx == len(pts) - 1)
            and ((v - pts[0]).Length < tol)
        ):
            obj.Closed = True
        # DNC: checks if point enter is equal to other, this could cause a OCC problem
        if v in pts:
            _err = translate("draft", "This object does not support possible " "coincident points")
            App.Console.PrintMessage(_err + "\n")
            return

        pts = self.recomputePointsBezier(obj, pts, node_idx, v, obj.Degree, moveTrackers=False)

        if obj.Closed:
            # check that the new point lies on the plane of the wire
            if hasattr(obj.Shape, "normalAt"):
                normal = obj.Shape.normalAt(0, 0)
                point_on_plane = obj.Shape.Vertexes[0].Point
                v.projectToPlane(point_on_plane, normal)
        pts[node_idx] = v
        obj.Points = pts

    def get_edit_point_context_menu(self, edit_command, obj, node_idx):
        return [
            (translate("draft", "Delete Point"), lambda: self.delete_point(obj, node_idx)),
            (translate("draft", "Make Sharp"), lambda: self.smoothBezPoint(obj, node_idx, "Sharp")),
            (
                translate("draft", "Make Tangent"),
                lambda: self.smoothBezPoint(obj, node_idx, "Tangent"),
            ),
            (
                translate("draft", "Make Symmetric"),
                lambda: self.smoothBezPoint(obj, node_idx, "Symmetric"),
            ),
        ]

    def get_edit_obj_context_menu(self, edit_command, obj, position):
        return [
            (translate("draft", "Add Point"), lambda: self.add_point(edit_command, obj, position)),
            (self.get_open_close_menu_text(obj), lambda: self.open_close_wire(obj)),
            (translate("draft", "Reverse Curve"), lambda: self.reverse_wire(obj)),
        ]

    def get_open_close_menu_text(self, obj):
        if obj.Closed:
            return translate("draft", "Open Curve")
        else:
            return translate("draft", "Close Curve")

    def init_preview_object(self, obj):
        return trackers.bezcurveTracker()

    def update_preview_object(self, edit_command, obj, node_idx, v):
        plist = edit_command.globalize_vectors(obj, obj.Points)
        pointList = self.recomputePointsBezier(
            obj, plist, node_idx, v, obj.Degree, moveTrackers=False
        )
        edit_command.ghost.update(pointList, obj.Degree)

    def recomputePointsBezier(self, obj, pts, idx, v, degree, moveTrackers=False):
        """
        (object, Points as list, nodeIndex as Int, App.Vector of new point, moveTrackers as Bool)
        return the new point list, applying the App.Vector to the given index point
        """
        # DNC: allows one to close the curve by placing ends close to each other
        tol = 0.001
        if (
            ((idx == 0) and ((v - pts[-1]).Length < tol))
            or (idx == len(pts) - 1)
            and ((v - pts[0]).Length < tol)
        ):
            obj.Closed = True
        # DNC: fix error message if edited point coincides with one of the existing points
        # if ( v in pts ) == False:
        knot = None
        ispole = idx % degree

        if ispole == 0:  # knot
            if degree >= 3:
                if idx >= 1:  # move left pole
                    knotidx = idx if idx < len(pts) else 0
                    pts[idx - 1] = pts[idx - 1] + v - pts[knotidx]
                    # if moveTrackers: # trackers are reset after editing
                    #    self.trackers[obj.Name][idx-1].set(pts[idx-1])
                if idx < len(pts) - 1:  # move right pole
                    pts[idx + 1] = pts[idx + 1] + v - pts[idx]
                    # if moveTrackers:
                    #    self.trackers[obj.Name][idx+1].set(pts[idx+1])
                if idx == 0 and obj.Closed:  # move last pole
                    pts[-1] = pts[-1] + v - pts[idx]
                    # if moveTrackers:
                    #    self.trackers[obj.Name][-1].set(pts[-1])

        elif ispole == 1 and (idx >= 2 or obj.Closed):  # right pole
            knot = idx - 1
            changep = idx - 2  # -1 in case of closed curve

        elif ispole == degree - 1 and idx <= len(pts) - 3:  # left pole
            knot = idx + 1
            changep = idx + 2

        elif ispole == degree - 1 and obj.Closed and idx == len(pts) - 1:  # last pole
            knot = 0
            changep = 1

        if knot is not None:  # we need to modify the opposite pole
            segment = int(knot / degree) - 1
            cont = obj.Continuity[segment] if len(obj.Continuity) > segment else 0
            if cont == 1:  # tangent
                pts[changep] = obj.Proxy.modifytangentpole(pts[knot], v, pts[changep])
                # if moveTrackers:
                #    self.trackers[obj.Name][changep].set(pts[changep])
            elif cont == 2:  # symmetric
                pts[changep] = obj.Proxy.modifysymmetricpole(pts[knot], v)
                # if moveTrackers:
                #    self.trackers[obj.Name][changep].set(pts[changep])
        pts[idx] = v

        return pts  # returns the list of new points, taking into account knot continuity

    def smoothBezPoint(self, obj, point, style="Symmetric"):
        """called when changing the continuity of a knot"""
        style2cont = {"Sharp": 0, "Tangent": 1, "Symmetric": 2}
        if point is None:
            return
        if not (utils.get_type(obj) == "BezCurve"):
            return
        pts = obj.Points
        deg = obj.Degree
        if deg < 2:
            return
        if point % deg != 0:  # point is a pole
            if deg >= 3:  # allow one to select poles
                if (point % deg == 1) and (point > 2 or obj.Closed):  # right pole
                    knot = point - 1
                    keepp = point
                    changep = point - 2
                elif point < len(pts) - 3 and point % deg == deg - 1:  # left pole
                    knot = point + 1
                    keepp = point
                    changep = point + 2
                elif point == len(pts) - 1 and obj.Closed:  # last pole
                    # if the curve is closed the last pole has the last
                    # index in the points lists
                    knot = 0
                    keepp = point
                    changep = 1
                else:
                    App.Console.PrintWarning(
                        translate("draft", "Cannot change knot belonging to pole %d" % point) + "\n"
                    )
                    return
                if knot:
                    if style == "Tangent":
                        pts[changep] = obj.Proxy.modifytangentpole(
                            pts[knot], pts[keepp], pts[changep]
                        )
                    elif style == "Symmetric":
                        pts[changep] = obj.Proxy.modifysymmetricpole(pts[knot], pts[keepp])
                    else:  # sharp
                        pass  #
            else:
                App.Console.PrintWarning(translate("draft", "Selection is not a knot") + "\n")
                return
        else:  # point is a knot
            if style == "Sharp":
                if obj.Closed and point == len(pts) - 1:
                    knot = 0
                else:
                    knot = point
            elif style == "Tangent" and point > 0 and point < len(pts) - 1:
                prev, next = obj.Proxy.tangentpoles(pts[point], pts[point - 1], pts[point + 1])
                pts[point - 1] = prev
                pts[point + 1] = next
                knot = point  # index for continuity
            elif style == "Symmetric" and point > 0 and point < len(pts) - 1:
                prev, next = obj.Proxy.symmetricpoles(pts[point], pts[point - 1], pts[point + 1])
                pts[point - 1] = prev
                pts[point + 1] = next
                knot = point  # index for continuity
            elif obj.Closed and (style == "Symmetric" or style == "Tangent"):
                if style == "Tangent":
                    pts[1], pts[-1] = obj.Proxy.tangentpoles(pts[0], pts[1], pts[-1])
                elif style == "Symmetric":
                    pts[1], pts[-1] = obj.Proxy.symmetricpoles(pts[0], pts[1], pts[-1])
                knot = 0
            else:
                App.Console.PrintWarning(
                    translate("draft", "Endpoint of Bézier curve cannot be smoothed") + "\n"
                )
                return
        segment = knot // deg  # segment index
        newcont = obj.Continuity[:]  # don't edit a property inplace !!!
        if not obj.Closed and (len(obj.Continuity) == segment - 1 or segment == 0):
            pass  # open curve
        elif (
            len(obj.Continuity) >= segment
            or obj.Closed
            and segment == 0
            and len(obj.Continuity) > 1
        ):
            newcont[segment - 1] = style2cont.get(style)
        else:  # should not happen
            App.Console.PrintWarning(
                "Continuity indexing error:"
                + "point:%d deg:%d len(cont):%d" % (knot, deg, len(obj.Continuity))
            )
        obj.Points = pts
        obj.Continuity = newcont

    def add_point(self, edit_command, obj, pos):
        """Add point to obj and reset trackers."""
        info, pt = edit_command.get_specific_object_info(obj, pos)
        if not info or (pt is None):
            return

        import Part

        pts = obj.Points
        if not info["Component"].startswith("Edge"):
            return  # clicked control point
        edgeindex = int(info["Component"].lstrip("Edge")) - 1
        wire = obj.Shape.Wires[0]
        bz = wire.Edges[edgeindex].Curve
        param = bz.parameter(pt)
        seg1 = wire.Edges[edgeindex].copy().Curve
        seg2 = wire.Edges[edgeindex].copy().Curve
        seg1.segment(seg1.FirstParameter, param)
        seg2.segment(param, seg2.LastParameter)
        if edgeindex == len(wire.Edges):
            # we hit the last segment, we need to fix the degree
            degree = wire.Edges[0].Curve.Degree
            seg1.increase(degree)
            seg2.increase(degree)
        edges = (
            wire.Edges[0:edgeindex]
            + [Part.Edge(seg1), Part.Edge(seg2)]
            + wire.Edges[edgeindex + 1 :]
        )
        pts = edges[0].Curve.getPoles()[0:1]
        for edge in edges:
            pts.extend(edge.Curve.getPoles()[1:])
        if obj.Closed:
            pts.pop()
        c = obj.Continuity
        # assume we have a tangent continuity for an arbitrarily split
        # segment, unless it's linear
        cont = 1 if (obj.Degree >= 2) else 0
        obj.Continuity = c[0:edgeindex] + [cont] + c[edgeindex:]

        obj.Points = pts

        obj.recompute()

    def delete_point(self, obj, node_idx):
        if len(obj.Points) <= 2:
            _err(translate("draft", "Active object must have more than two points/nodes"))
            return

        pts = obj.Points
        pts.pop(node_idx)
        obj.Points = pts
        obj.Proxy.resetcontinuity(obj)
        obj.recompute()

    def open_close_wire(self, obj):
        obj.Closed = not obj.Closed
        obj.recompute()

    def reverse_wire(self, obj):
        obj.Points = reversed(obj.Points)
        obj.recompute()


## @}
